#-*- coding:utf-8 -*-
#
# generation/connect_algorithms.py
#
# This file is part of the NNGT project, a graph-library for standardized and
# and reproducible graph analysis: generate and analyze networks with your
# favorite graph library (graph-tool/igraph/networkx) on any platform, without
# any change to your code.
# Copyright (C) 2015-2021 Tanguy Fardet
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

""" Generation tools for NNGT """

import logging
import numpy as np
import scipy.sparse as ssp
from numpy.random import randint

import nngt
from nngt.lib import InvalidArgument
from nngt.lib.connect_tools import *


__all__ = [
    "_all_to_all",
    "_circular",
    "_distance_rule",
    "_erdos_renyi",
    "_fixed_degree",
    "_from_degree_list",
    "_gaussian_degree",
    "_newman_watts",
    "_price_scale_free",
    "_random_scale_free",
    "_unique_rows",
    "_watts_strogatz",
]

logger = logging.getLogger(__name__)

MAXTESTS = 1000 # ensure that generation will finish
EPS = 0.00001


# ---------------------- #
# Graph model generation #
# ---------------------- #

def _all_to_all(source_ids, target_ids, directed=True, multigraph=False,
                distance=None, **kwargs):
    num_sources, num_targets = len(source_ids), len(target_ids)

    # find common nodes
    edges  = None
    common = set(source_ids).intersection(target_ids)

    num_edges = int(0.5*(1 + directed)*num_sources*num_targets)

    if common:
        num_edges -= int(0.5*(1 + directed)*len(common))

        edges = np.full((num_edges, 2), -1, dtype=int)

        current_edges = 0
        next_enum     = 0

        for s in source_ids:
            if s in common:
                if directed:
                    next_enum = current_edges + num_targets - 1

                    edges[current_edges:next_enum, 1] = \
                        target_ids[target_ids != s]
                else:
                    tgts = [t for t in target_ids if t not in common or t > s]

                    next_enum = current_edges + len(tgts)

                    edges[current_edges:next_enum, 1] = tgts

                edges[current_edges:next_enum, 0] = s

                current_edges = next_enum
            else:
                next_enum = current_edges + num_targets
                edges[current_edges:next_enum, 0] = s
                edges[current_edges:next_enum, 1] = target_ids
                current_edges = next_enum
    else:
        edges       = np.empty((num_edges, 2), dtype=int)
        edges[:, 0] = np.repeat(source_ids, num_targets)
        edges[:, 1] = np.tile(target_ids, num_sources)

    if distance is not None and 'positions' in kwargs:
        pos       = kwargs['positions']
        x, y      = pos[0], pos[1]
        vectors   = np.array((x[edges[:, 1]] - x[edges[:, 0]],
                              y[edges[:, 1]] - y[edges[:, 0]]))
        distance.extend(np.linalg.norm(vectors, axis=0))

    return edges


def _total_degree_list(source_ids, target_ids, degree_list, directed=True,
                       multigraph=False, **kwargs):
    ''' Called from _from_degree_list '''
    degree_list = np.array(degree_list, dtype=int)

    num_source, num_target = len(source_ids), len(target_ids)
    edges = int(0.5*np.sum(degree_list))

    # check if the sequence is not obviously impossible
    if not multigraph:
        assert np.all(degree_list <= (num_target - 1)), \
            "Some degrees are higher than the number of available targets."

    if np.sum(degree_list) % 2 != 0:
        raise ValueError("The sum of the degrees must ben even.")

    # check one pop
    b_one_pop, source_set, target_set = _check_num_edges(
        source_ids, target_ids, edges, directed, multigraph, return_sets=True)

    ecurrent = 0
    edges_hash = set()
    recip_hash = set()

    ia_edges = np.full((edges, 2), -1, dtype=int)

    rng = nngt._rng

    if b_one_pop:
        # check that the degree sequence is graphic (i.e. can be realized)
        sorted_degrees = np.sort(degree_list)[::-1]
        for k in range(1, num_source + 1):
            partial_sum = np.sum(sorted_degrees[:k])
            capped_deg  = np.minimum(sorted_degrees[k:], k)

            if not partial_sum <= k*(k-1) + np.sum(capped_deg):
                raise ValueError("The degree sequence provided is not "
                                 "graphical and cannot be realized.")

        num_tests = 0

        while ecurrent < edges and num_tests < MAXTESTS:
            # initial procedure
            remaining = max(int(0.5*(edges - ecurrent)), 1)

            data = np.repeat(source_ids, degree_list)
            sources = rng.choice(data, remaining, replace=False)

            deg_tmp = degree_list.copy()
            np.add.at(deg_tmp, sources, -1)

            data = np.repeat(source_ids, deg_tmp)
            targets = rng.choice(data, remaining, replace=False)

            new_edges = np.array((sources, targets), dtype=int).T

            ia_edges, new_etot =  _filter(
                ia_edges, new_edges, ecurrent, edges_hash, b_one_pop,
                multigraph, directed=directed, recip_hash=recip_hash)

            np.add.at(degree_list, ia_edges[ecurrent:new_etot].ravel(), -1)

            ecurrent = new_etot

            # trying to correct things if initial procedure did not converge
            if num_tests >= 0.5*MAXTESTS:
                nnz_degree = degree_list != 0
                unfinished = np.where(nnz_degree)[0]
                num_choice = max(int(0.5*len(unfinished)), 1)

                # pick random edges
                ids    = set(rng.choice(ecurrent, num_choice, replace=False))
                chosen = np.array([ia_edges[i] for i in ids], dtype=int)

                # if unfinished is odd then we need to readd a node that has
                # a remaining degree of 2
                if len(unfinished) % 2 == 1:
                    even_deg = np.where((degree_list % 2 == 0)*nnz_degree)[0]
                    add_node = rng.choice(even_deg, 1)
                    unfinished = list(unfinished) + list(add_node)

                # try to pair them differently
                rng.shuffle(unfinished)

                new_edges = np.array(
                    [(u, v) for u, v in zip(unfinished, chosen.ravel())],
                    dtype=int)

                # randomize direction if directed
                if directed:
                    num_reverse = rng.binomial(2*num_choice, 0.5)
                    reverse     = rng.choice(2*num_choice, num_reverse,
                                             replace=False)
                    new_edges[reverse, 0], new_edges[reverse, 1] = \
                        new_edges[reverse, 1], new_edges[reverse, 0]

                # check that new_edges are indeed new
                skip = False

                for e in new_edges:
                    if e[0] == e[1]:
                        skip = True
                        break
                        
                    if tuple(e) in edges_hash:
                        skip = True
                        break

                    if not directed and tuple(e) in recip_hash:
                        skip = True
                        break

                if skip:
                    num_tests += 1
                    continue

                # remove old ones from edges_hash
                edges_hash -= {tuple(e) for e in chosen}

                if not directed:
                    recip_hash -= {tuple(e) for e in chosen[:, ::-1]}

                # remove chosen edges from existing edges
                ia_edges[:ecurrent - num_choice] = np.array(
                    [e for i, e in enumerate(ia_edges[:ecurrent])
                     if i not in ids])

                ecurrent -= num_choice

                ia_edges, new_etot =  _filter(
                    ia_edges, new_edges, ecurrent, edges_hash, b_one_pop,
                    multigraph, directed=directed, recip_hash=recip_hash)

                decr  = list(new_edges.ravel())
                incr  = list(chosen.ravel())

                np.add.at(degree_list, incr, 1)
                np.add.at(degree_list, decr, -1)
                
                ecurrent += len(new_edges)

            num_tests += 1

        if num_tests == MAXTESTS:
            raise RuntimeError("Graph generation did not converge.")

        return ia_edges

    raise NotImplementedError("not available if sources != targets.")
    

def _from_degree_list(source_ids, target_ids, degree_list, degree_type="in",
                      directed=True, multigraph=False, existing_edges=None,
                      **kwargs):
    assert len(degree_list) == len(source_ids), \
        "One degree per source neuron must be provided."

    source_ids = np.array(source_ids).astype(int)
    target_ids = np.array(target_ids).astype(int)
    num_source, num_target = len(source_ids), len(target_ids)

    # type of degree
    degree_type = _set_degree_type(degree_type)

    b_out = (degree_type == "out")
    b_total = (degree_type == "total")

    # directed case for in/out-degree
    if directed and not b_total:
        edges = np.sum(degree_list)

        b_one_pop = _check_num_edges(
            source_ids, target_ids, edges, directed, multigraph)

        num_einit  = 0 if existing_edges is None else existing_edges.shape[0]
        num_etotal = num_einit

        ia_edges = np.full((num_einit + edges, 2), -1, dtype=int)

        if num_einit:
            ia_edges[:num_einit, :] = existing_edges

        idx = 0 if b_out else 1  # differenciate source / target

        # keep track of degrees for undirected case
        target_degrees  = {}
        current_targets = {u: [] for u in source_ids}

        for u, d in zip(source_ids, degree_list):
            if existing_edges is not None:
                current_targets[u].extend(
                    existing_edges[existing_edges[:, idx] == u, 1 - idx])

            if not multigraph and (num_target - len(current_targets[u]) < d):
                raise RuntimeError(
                    ("Node {node} already has a degree {d0} which means that "
                     "there are not enough targets left to add {d} edges."
                    ).format(node=u, d0=len(current_targets[u]), target=d))

            target_degrees[u] = d

        rng = nngt._rng

        # create new connections
        for v, degree_v in target_degrees.items():
            existing_v = current_targets[v]
            targets_v  = [] # where the targets will be stored

            # set targets
            rm = np.where(target_ids == v)[0]
            rm = rm[0] if len(rm) else -1

            tgts = (target_ids if rm == -1 else
                    np.concatenate((target_ids[:rm], target_ids[rm+1:])))

            # select targets
            while len(targets_v) < degree_v:
                var = rng.choice(tgts, degree_v - len(targets_v),
                                 replace=multigraph)

                targets_v.extend(var)

                if not multigraph:
                    s = set(existing_v + targets_v)

                    targets_v  = list(s.difference(existing_v))
                    existing_v = list(s)


            ia_edges[num_etotal:num_etotal + degree_v, idx] = v
            ia_edges[num_etotal:num_etotal + degree_v, 1 - idx] = targets_v

            num_etotal += degree_v

        return ia_edges[num_einit:]

    # unidrected or total-degree cases
    return _total_degree_list(source_ids, target_ids, degree_list,
                              directed=directed, multigraph=multigraph)


def _fixed_degree(source_ids, target_ids, degree, degree_type="in",
                  reciprocity=-1, directed=True, multigraph=False,
                  existing_edges=None, **kwargs):
    degree = int(degree)
    assert degree >= 0, "A positive value is required for `degree`."

    num_source = len(source_ids)

    # edges (we set the in, out, or total degree of the source neurons)
    lst_deg = np.full(num_source, degree, dtype=int)

    return _from_degree_list(
        source_ids, target_ids, lst_deg, degree_type=degree_type,
        directed=directed, multigraph=multigraph,
        existing_edges=existing_edges, **kwargs)


def _gaussian_degree(source_ids, target_ids, avg, std, degree_type="in",
                     directed=True, multigraph=False, existing_edges=None,
                     **kwargs):
    ''' Connect nodes with a Gaussian distribution '''
    # switch values to float
    avg = float(avg)
    std = float(std)

    assert avg >= 0, "A positive value is required for `avg`."
    assert std >= 0, "A positive value is required for `std`."

    num_source = len(source_ids)

    # type of degree
    degree_type = _set_degree_type(degree_type)

    b_total = (degree_type == "total")

    # edges (we set the in, out, or total degree of the source neurons)
    rng = nngt._rng

    degrees = np.around(
        np.maximum(rng.normal(avg, std, num_source), 0.)).astype(int)

    if b_total or not directed:
        # check that the sum of the degrees is even
        if np.sum(degrees) % 2 != 0:
            idx = -1

            # correct if its not the case
            while idx < 0 or degrees[idx] == 0:
                idx = rng.choice(num_source)
                if degrees[idx] > 0:
                    degrees[idx] -= 1

    return _from_degree_list(
        source_ids, target_ids, degrees, degree_type=degree_type,
        directed=directed, multigraph=multigraph,
        existing_edges=existing_edges, **kwargs)


def _random_scale_free(source_ids, target_ids, in_exp, out_exp, density=None,
                       edges=None, avg_deg=None, reciprocity=-1,
                       directed=True, multigraph=False, **kwargs):
    ''' Connect the nodes with power law distributions '''
    source_ids = np.array(source_ids).astype(int)
    target_ids = np.array(target_ids).astype(int)
    num_source, num_target = len(source_ids), len(target_ids)
    edges, pre_recip_edges = _compute_connections(num_source, num_target,
                                density, edges, avg_deg, directed, reciprocity)
    b_one_pop = _check_num_edges(
        source_ids, target_ids, edges, directed, multigraph)

    arr_edges = np.full((edges, 2), -1, dtype=int)
    num_ecurrent, num_test = 0, 0
    edges_hash = set()
    recip_hash = None if not directed else set()

    # lists containing the in/out-degrees for all nodes
    in_deg = np.random.pareto(in_exp, num_target) + 1
    out_deg = np.random.pareto(out_exp, num_source) + 1

    sum_in, sum_out = np.sum(in_deg), np.sum(out_deg)

    in_deg = np.around(np.multiply(
        pre_recip_edges / sum_in, in_deg)).astype(int)
    out_deg = np.around(np.multiply(
        pre_recip_edges / sum_out, out_deg)).astype(int)

    sum_in, sum_out = np.sum(in_deg), np.sum(out_deg)

    # correct
    while sum_in != pre_recip_edges or sum_out != pre_recip_edges:
        diff_in  = sum_in - pre_recip_edges
        diff_out = sum_out - pre_recip_edges

        idx_correct_in = randint(0, num_target, np.abs(diff_in))
        idx_correct_out = randint(0, num_source, np.abs(diff_out))
        in_deg[idx_correct_in] -= 1*np.sign(diff_in)
        out_deg[idx_correct_out] -= 1*np.sign(diff_out)
        in_deg[in_deg < 0] = 0
        out_deg[out_deg < 0] = 0
        sum_in, sum_out = np.sum(in_deg), np.sum(out_deg)

    # make the edges
    sources = np.repeat(source_ids, out_deg)
    targets = np.repeat(target_ids, in_deg)

    np.random.shuffle(targets)

    arr_edges_tmp = np.array((sources, targets)).T

    arr_edges, num_ecurrent = _filter(arr_edges, arr_edges_tmp, num_ecurrent,
                                     edges_hash, b_one_pop, multigraph,
                                     directed=directed, recip_hash=recip_hash)

    while num_ecurrent != pre_recip_edges and num_test < MAXTESTS:
        num_desired = pre_recip_edges-num_ecurrent
        sources_tmp = np.random.choice(sources, num_desired)
        targets_tmp = np.random.choice(targets, num_desired)
        arr_edges_tmp = np.array([sources_tmp, targets_tmp]).T
        arr_edges, num_ecurrent = _filter(
            arr_edges, arr_edges_tmp, num_ecurrent, edges_hash, b_one_pop,
            multigraph, directed=directed, recip_hash=recip_hash)
        num_test += 1

    if directed and reciprocity > 0:
        while num_ecurrent != edges and num_test < MAXTESTS:
            keep = np.random.choice(num_ecurrent, edges-num_ecurrent,
                                    replace=multigraph)
            arr_edges[num_ecurrent:] = arr_edges[keep]
            num_ecurrent = edges
            if not multigraph:
                arr_edges_tmp = _unique_rows(arr_edges)
                num_ecurrent = arr_edges_tmp.shape[0]
                arr_edges[:num_ecurrent,:] = arr_edges_tmp
            num_test += 1

    return arr_edges


def _erdos_renyi(source_ids, target_ids, density=None, edges=None,
                 avg_deg=None, reciprocity=-1, directed=True, multigraph=False,
                 **kwargs):
    '''
    Returns a numpy array of dimension (2,edges) that describes the edge list
    of an Erdos-Renyi graph.
    @todo: perform all the calculations here
    '''
    source_ids = np.array(source_ids).astype(int)
    target_ids = np.array(target_ids).astype(int)
    num_source, num_target = len(source_ids), len(target_ids)
    edges, pre_recip_edges = _compute_connections(num_source, num_target,
                                density, edges, avg_deg, directed, reciprocity)

    b_one_pop = _check_num_edges(
        source_ids, target_ids, edges, directed, multigraph)

    ia_edges = np.full((edges, 2), -1, dtype=int)
    num_test, num_ecurrent = 0, 0 # number of tests and current number of edges
    edges_hash = set()
    recip_hash = None if directed else set()

    while num_ecurrent != pre_recip_edges and num_test < MAXTESTS:
        ia_sources = source_ids[randint(0, num_source,
                                        pre_recip_edges-num_ecurrent)]
        ia_targets = target_ids[randint(0, num_target,
                                        pre_recip_edges-num_ecurrent)]
        ia_edges_tmp = np.array([ia_sources,ia_targets]).T
        ia_edges, num_ecurrent = _filter(
            ia_edges, ia_edges_tmp, num_ecurrent, edges_hash, b_one_pop,
            multigraph, directed=directed, recip_hash=recip_hash)
        num_test += 1

    if directed and reciprocity > 0:
        while num_ecurrent != edges and num_test < MAXTESTS:
            ia_indices = randint(0, pre_recip_edges,
                                           edges-num_ecurrent)
            ia_edges[num_ecurrent:] = ia_edges[ia_indices]
            num_ecurrent = edges
            if not multigraph:
                ia_edges_tmp = _unique_rows(ia_edges)
                num_ecurrent = ia_edges_tmp.shape[0]
                ia_edges[:num_ecurrent,:] = ia_edges_tmp
            num_test += 1

    return ia_edges


def _price_scale_free(ids, m, c, gamma, reciprocity, directed, multigraph,
                      **kwargs):
    '''
    Generate a Price network.
    '''
    ids = np.array(ids).astype(int)

    num_nodes = len(ids)

    assert isinstance(m, int) and m >= 0, "`m` must be a positive integer."

    if directed:
        assert c > 0, "`c` > 0 is required for directed graphs."
    else:
        assert c > -1, "`c` > -1 is required for undirected graphs."

    assert 0 <= reciprocity <= 1, "`reciprocity` must be in [0, 1]."

    in_degrees = np.zeros(num_nodes)

    edges = []
    edges_hash = set()
    recip_hash = None if directed else set()

    rng = nngt._rng

    for i, n in zip(range(1, num_nodes), ids[1:]):
        proba = None if i == 1 else np.power(in_degrees[:i], gamma) + c

        if proba is not None:
            proba /= proba.sum()

        edges_i = []

        m_i = min(i, m)

        while len(edges_i) < m_i:
            targets = rng.choice(i, size=m_i, replace=False, p=proba)

            for t in targets:
                e = (n, ids[t])

                if e not in edges_hash:
                    if directed:
                        edges_i.append(e)
                        in_degrees[t] += 1
                    elif e not in recip_hash:
                        edges_i.append(e)
                        in_degrees[i] += 1
                        in_degrees[t] += 1

        edges.extend(edges_i)

        if directed and reciprocity > 0:
            make_recip = rng.random(m) < reciprocity

            for b, e in zip(make_recip, edges_i):
                if b:
                    edges.append(e[::-1])
                    in_degrees[i] += 1

    return edges


def _circular(source_ids, target_ids, coord_nb, reciprocity=1, directed=True,
              reciprocity_choice="random", **kwargs):
    '''
    Circular graph.

    Note
    ----
    `source_ids` and `target_ids` are only there for compatibility with the
    connect functions, this algorithm requires a single population.
    This check (if necessary) is performed above.
    '''
    if reciprocity_choice not in ("random", "closest", "closest-ordered"):
        # note: only "random" and "closest" are publicly advertised,
        # "closest-ordered" is only for internal use in lattice_rewire.
        raise ValueError("Valid entries for `reciprocity_choice` are "
                         "'random' and 'closest'.")

    if coord_nb % 2:
        raise ValueError("`coord_nb` must be even.")

    if reciprocity == 1 or not directed:
        return _circular_full(source_ids, coord_nb, directed)
    elif directed:
        return _circular_directed_recip(source_ids, coord_nb, reciprocity,
                                        reciprocity_choice=reciprocity_choice)

    raise ValueError("For undirected graphs, reciprocity is 1 by definition.")


def _circular_directed_recip(node_ids, coord_nb, reciprocity,
                             reciprocity_choice="random", **kwargs):
    ''' Circular graph with given reciprocity '''
    nodes    = len(node_ids)
    edges    = int(np.round(0.5 * nodes * coord_nb *
                   (1 + reciprocity / (2 - reciprocity))))
    init_deg = int(0.5*coord_nb)

    # sources and targets
    sources = np.full(edges, -1, dtype=int)
    targets = np.full(edges, -1, dtype=int)

    # set non-reciprocal edges using the full undirected circular graph
    init_edges = _circular_full(node_ids, coord_nb, False)
    num_init   = len(init_edges)

    sources[:num_init] = init_edges[:, 0]
    targets[:num_init] = init_edges[:, 1]

    # then we randomize the direction of these E_init edges
    # this is equivalent to reversing E edges with E from Binom(E_init, 0.5)
    rng = nngt._rng
    E   = rng.binomial(num_init, 0.5)

    chosen = rng.choice(num_init, E, replace=False)

    sources[chosen], targets[chosen] = targets[chosen], sources[chosen]

    # set reciprocal edges
    num_recip = edges - nodes*init_deg

    if num_recip:
        if reciprocity_choice == "random":
            chosen = rng.choice(nodes*init_deg, size=num_recip, replace=False)

            sources[-num_recip:] = targets[chosen]
            targets[-num_recip:] = sources[chosen]
        elif reciprocity_choice == "closest":
            # closest connections are the first ones, so if
            # num_recip = k*nodes + l then we reverse all first k*nodes
            # then we randomly chose the remaining l
            remainder = num_recip % nodes
            rounds    = num_recip - remainder
            
            sources[-num_recip:-num_recip + rounds] = targets[:rounds]
            targets[-num_recip:-num_recip + rounds] = sources[:rounds]

            # chose randomly the remaining connections
            stop = min(rounds + nodes, edges)
            chosen = rng.choice([i for i in range(rounds, stop)],
                                size=remainder, replace=False)

            sources[-num_recip + rounds:] = targets[chosen]
            targets[-num_recip + rounds:] = sources[chosen]
        else:
            # deterministic (ordered) closest connections, only for the
            # lattice rewiring (see function _lattice_shuffle_eattr)
            # closest connections are the first ones
            remainder = num_recip % nodes
            rounds    = num_recip - remainder

            start = edges - num_recip
            stop  = edges - num_recip + rounds

            sources[start:stop] = targets[:rounds]
            targets[start:stop] = sources[:rounds]

            # chose randomly the remaining connections
            start = stop
            stop  = min(rounds + nodes, edges)

            chosen = rng.choice(list(range(rounds, stop)), size=remainder,
                                replace=False)

            sources[start:] = targets[chosen]
            targets[start:] = sources[chosen]

    return np.array([sources, targets], dtype=int).T


def _circular_full(node_ids, coord_nb, directed, **kwargs):
    ''' Create a circular graph with all possible edges '''
    nodes = len(node_ids)

    dist = int(0.5*coord_nb)

    out_deg = coord_nb if directed else dist

    # create the graph using a continuous range from zero
    sources = np.tile(np.arange(0, nodes).astype(int), out_deg)

    # prepare conversion dict if nodes do not start from zero or are not
    # contiguous. Initial generation assumes continuous range from zero
    convertor = {i: n for i, n in enumerate(node_ids)}

    # create the connection mask
    start = -dist if directed else 0
    stop  = dist + 1

    conn_mask = np.concatenate((np.arange(start, 0), np.arange(1, stop)))

    # make the targets and put them back into [0, nodes - 1]
    targets   = np.repeat(conn_mask, nodes) + sources

    targets[targets < 0] += nodes
    targets[targets >= nodes] -= nodes

    # convert back to ids
    sources = np.array([convertor[i] for i in sources])
    targets = np.array([convertor[i] for i in targets])

    return np.array((sources, targets), dtype=int).T


def _newman_watts(source_ids, target_ids, coord_nb, proba_shortcut,
                  reciprocity_circular=1, reciprocity_choice_circular="random",
                  edges=None, directed=True, multigraph=False, **kwargs):
    '''
    Returns a numpy array of dimension (num_edges,2) that describes the edge
    list of a Newman-Watts graph.

    Note
    ----
    `source_ids` and `target_ids` are only there for compatibility with the
    connect functions, this algorithm requires a single population.
    This check (if necessary) is performed above.
    '''
    nodes      = len(source_ids)
    source_ids = np.array(source_ids, dtype=int)
    target_ids = np.array(target_ids, dtype=int)

    # check the number of edges
    direct_factor  = 0.5*(1 + directed)
    recip_factor   = 1 + reciprocity_circular / (2 - reciprocity_circular)
    circular_edges = int(np.round(
        0.5 * direct_factor * nodes * coord_nb * recip_factor))

    if edges is None:
        rng   = nngt._rng
        edges = circular_edges + rng.binomial(circular_edges, proba_shortcut)
    elif edges < circular_edges:
        raise ValueError("`edges` must be greater or equal to "
                         "{} given current arguments.".format(circular_edges))

    b_one_pop = _check_num_edges(
        source_ids, target_ids, edges, directed, multigraph)

    if not b_one_pop:
        raise InvalidArgument("This graph model can only be used if source "
                              "and target populations are the same.")

    # generate the initial circular graph
    ia_edges = np.full((edges, 2), -1, dtype=int)

    ia_edges[:circular_edges, :] = _circular(
        source_ids, source_ids, coord_nb, reciprocity_circular, directed,
        reciprocity_choice=reciprocity_choice_circular)

    # add the random connections
    num_test, num_ecurrent = 0, circular_edges
    edges_hash = set(tuple(e) for e in ia_edges[:circular_edges])
    recip_hash = None

    if not directed:
        recip_hash = set((tuple(e) for e in ia_edges[:circular_edges, ::-1]))

    rng = nngt._rng

    while num_ecurrent != edges and num_test < MAXTESTS:
        todo   = edges - num_ecurrent
        chosen = rng.choice(source_ids, int(2*todo))

        ia_edges, num_ecurrent = _filter(
            ia_edges, chosen.reshape(todo, 2), num_ecurrent, edges_hash,
            b_one_pop, multigraph, directed=directed, recip_hash=recip_hash)

        num_test += 1

    return ia_edges


def _watts_strogatz(
        source_ids, target_ids, coord_nb, proba_shortcut,
        reciprocity_circular=1, reciprocity_choice_circular="random",
        shuffle="random", directed=True, multigraph=False, **kwargs):
    '''
    Returns a numpy array of dimension (num_edges,2) that describes the edge
    list of a Watts-Strogatz graph.

    Note
    ----
    `source_ids` and `target_ids` are only there for compatibility with the
    connect functions, this algorithm requires a single population.
    This check (if necessary) is performed above.
    '''
    assert shuffle in ('sources', 'targets', 'random'), \
        "Shuffle must be either 'sources', 'targets', or 'random'."

    nodes      = len(source_ids)
    source_ids = np.array(source_ids, dtype=int)
    target_ids = np.array(target_ids, dtype=int)

    # check the number of edges
    direct_factor  = 0.5*(1 + directed)
    recip_factor   = 0.5*(1 + reciprocity_circular)

    deg   = int(coord_nb * recip_factor * direct_factor)
    edges = nodes * deg

    b_one_pop = _check_num_edges(
        source_ids, target_ids, edges, directed, multigraph)

    if not b_one_pop:
        raise InvalidArgument("This graph model can only be used if source "
                              "and target populations are the same.")

    # generate the initial circular graph
    ia_edges = _circular(
        source_ids, source_ids, coord_nb, reciprocity_circular, directed,
        reciprocity_choice=reciprocity_choice_circular)

    # randomize some of the outgoing connections
    node_set = set(source_ids)

    rng = nngt._rng

    edge_hash = set(tuple(e) for e in ia_edges)

    if not directed:
        edge_hash.update([tuple(e) for e in ia_edges[:, ::-1]])

    # compute how many we rewire in total, choose them
    rewire = rng.binomial(edges, proba_shortcut)
    chosen = rng.choice(edges, rewire, replace=False).tolist()

    # rewire
    keep = [0]*rewire

    if shuffle == "random":
        keep = rng.integers(1, size=rewire, endpoint=True).tolist()
    elif shuffle == "sources":
        keep = [1]*rewire

    new_targets = rng.integers(nodes, size=rewire).tolist()

    max_tests   = 50*rewire
    num_test    = 0
    num_rewired = 0

    while num_rewired < rewire and num_test < max_tests:
        # if used all previous chosen, regenerate some
        if not chosen:
            remaining = rewire - num_rewired

            chosen = rng.choice(edges, remaining, replace=False).tolist()

            keep = [0]*remaining
            if shuffle == "random":
                keep = rng.integers(1, size=remaining, endpoint=True).tolist()
            elif shuffle == "sources":
                keep = [1]*remaining

            new_targets = rng.integers(nodes, size=remaining).tolist()

        # try to rewire
        idx1 = chosen.pop()
        idx2 = keep.pop()

        old_edge = tuple(ia_edges[idx1])

        s = old_edge[idx2]
        t = new_targets.pop()
        e = (s, t) if idx2 == 0 else (t, s)

        mtests = 50*nodes
        ntest  = 0

        while (s == t or e in edge_hash) and ntest < mtests:
            t = rng.integers(nodes)
            e = (s, t) if idx2 == 0 else (t, s)
            ntest += 1

        # update test here to count skips
        num_test += 1

        # chosen node may already reached max degree, skip
        if ntest == mtests:
            continue

        # rewire edge if successful
        ia_edges[idx1, 1 - idx2] = t

        # add new edge and remove old one
        edge_hash.add(e)

        edge_hash -= {old_edge}

        if not directed:
            edge_hash.add(e[::-1])

            edge_hash -= {old_edge[::-1]}

        num_rewired += 1

    if max_tests and num_test == max_tests:
        raise RuntimeError("Algorithm did not converge.")

    return ia_edges


def _distance_rule(source_ids, target_ids, density=None, edges=None,
                   avg_deg=None, scale=-1, rule="exp", max_proba=-1,
                   shape=None, positions=None, directed=True, multigraph=False,
                   distance=None, **kwargs):
    '''
    Returns a distance-rule graph
    '''
    distance = [] if distance is None else distance
    edges_hash = set()
    recip_hash = None if directed else set()

    # compute the required values
    source_ids = np.array(source_ids).astype(int)
    target_ids = np.array(target_ids).astype(int)

    num_source, num_target = len(source_ids), len(target_ids)
    num_edges = 0

    if max_proba <= 0:
        num_edges, _ = _compute_connections(
            num_source, num_target, density, edges, avg_deg, directed,
            reciprocity=-1)

    b_one_pop = _check_num_edges(
        source_ids, target_ids, num_edges, directed, multigraph)

    num_neurons = len(set(np.concatenate((source_ids, target_ids))))

    # for each node, check the neighbours that are in an area where
    # connections can be made: +/- scale for lin, +/- 10*scale for exp.
    # Get the sources and associated targets for each MPI process
    list_targets = []
    lim = scale if rule == 'lin' else 10*scale

    for s in source_ids:
        keep  = (np.abs(positions[0, target_ids] - positions[0, s]) < lim)
        keep *= (np.abs(positions[1, target_ids] - positions[1, s]) < lim)
        if b_one_pop:
            idx = np.where(target_ids == s)[0][0]
            keep[idx] = 0
        list_targets.append(target_ids[keep])

    # the number of trials should be done depending on the number of
    # neighbours that each node has, so compute this number
    tot_neighbours = 0
    neigh_norm     = 1.

    for tgt_list in list_targets:
        tot_neighbours += len(tgt_list)

    if max_proba <= 0:
        assert tot_neighbours > num_edges, \
            "Scale is too small: there are not enough close neighbours to " +\
            "create the required number of connections. Increase `scale` " +\
            "or `neuron_density`."

        neigh_norm = 1. / tot_neighbours

    # try to create edges until num_edges is attained
    ia_edges = None
    num_ecurrent = 0

    if max_proba <= 0:
        ia_edges = np.full((num_edges, 2), -1, dtype=int)
        while num_ecurrent < num_edges:
            trials = []
            for tgt_list in list_targets:
                trials.append(max(
                    int(len(tgt_list)*(num_edges - num_ecurrent)*neigh_norm),
                    1))
            edges_tmp = [[], []]
            dist = []
            dist_tmp = []
            total_trials = int(np.sum(trials))
            local_targets = []
            local_sources = []
            current_pos = 0
            for s, tgts, num_try in zip(source_ids, list_targets, trials):
                if len(tgts):
                    t = np.random.randint(0, len(tgts), num_try)
                    local_targets.extend(tgts[t])
                    local_sources.extend((s for _ in range(num_try)))
            local_targets = np.array(local_targets, dtype=int)
            local_sources = np.array(local_sources, dtype=int)
            test = dist_rule(rule, scale, positions[:, local_sources],
                             positions[:, local_targets], dist=dist_tmp)
            test = np.greater(test, np.random.uniform(size=len(test)))
            edges_tmp[0].extend(local_sources[test])
            edges_tmp[1].extend(local_targets[test])
            dist = np.array(dist_tmp)[test]

            edges_tmp = np.array(edges_tmp).T

            # assess the current number of edges
            # if we're at the end, we'll make too many edges, so we keep only
            # the necessary fraction that we pick randomly
            num_desired = num_edges - num_ecurrent
            if num_desired < len(edges_tmp):
                chosen = {}
                while len(chosen) != num_desired:
                    idx = np.random.randint(
                        0, len(edges_tmp), num_desired - len(chosen))
                    for i in idx:
                        chosen[i] = None
                idx = list(chosen.keys())
                edges_tmp = edges_tmp[idx]
                dist = dist[idx]

            ia_edges, num_ecurrent = _filter(
                ia_edges, edges_tmp, num_ecurrent, edges_hash, b_one_pop,
                multigraph, directed=directed, recip_hash=recip_hash,
                distance=distance, dist_tmp=dist)
    else:
        sources, targets = [], []
        for i, s in enumerate(source_ids):
            local_tgts = np.array(list_targets[i], dtype=int)
            if len(local_tgts):
                dist_tmp = []
                test = max_proba_dist_rule(
                    rule, scale, max_proba, positions[:, s],
                    positions[:, local_tgts], dist=dist_tmp)
                test = np.greater(test, np.random.uniform(size=len(test)))
                added = np.sum(test)
                sources.extend((s for _ in range(added)))
                targets.extend(local_tgts[test])
                distance.extend(np.array(dist_tmp)[test])
        ia_edges = np.array([sources, targets]).T

    return ia_edges



#!/usr/bin/env python
#-*- coding:utf-8 -*-

# test_nest.py

# This file is part of the NNGT module
# Distributed as a free software, in the hope that it will be useful, under the
# terms of the GNU General Public License.

import nngt
import nngt.generation as ng
import numpy as np
import pytest


def make_nest_net(size, w, deg):
    '''
    Create a network in NEST
    '''
    import nngt.simulation as ns

    net = nngt.Network.exc_and_inhib(size)

    pop = net.population

    ng.connect_groups(net, pop, pop, graph_model="fixed_degree",
                      degree=deg, degree_type="out", weights=w)

    gids = net.to_nest()

    return net, gids


@pytest.mark.skipif(nngt.get_config('mpi'), reason="Don't test for MPI")
def test_net_creation():
    '''
    Test the creation of a network in NEST.
    '''
    nest = pytest.importorskip("nest")

    nest.ResetKernel()

    w = 5.

    net, gids = make_nest_net(100, w, deg=10)

    # check nodes and connections
    assert len(gids) == net.node_nb()

    conn = nest.GetConnections()

    assert len(conn) == net.edge_nb()

    weights = np.array([d['weight'] for d in nest.GetStatus(conn)])

    # check inhibitory connections
    etypes = net.get_edge_types()

    num_i = np.sum(etypes == -1)

    assert num_i == int(0.2*net.edge_nb())

    assert np.sum(weights == -w) == num_i

    assert set(weights) == {-w, w}


@pytest.mark.skipif(nngt.get_config('mpi'), reason="Don't test for MPI")
def test_utils():
    '''
    Check NEST utility functions
    '''
    nest = pytest.importorskip("nest")

    import nngt.simulation as ns

    nest.ResetKernel()

    resol = nest.GetKernelStatus('resolution')

    w = 5.

    net, gids = make_nest_net(100, w, deg=10)

    # set inputs
    ns.set_noise(gids, 0., 20.)

    assert len(nest.GetConnections()) == net.edge_nb() + net.node_nb()

    ns.set_poisson_input(gids, 5.)

    assert len(nest.GetConnections()) == net.edge_nb() + 2*net.node_nb()

    ns.set_minis(net, base_rate=1.5, weight=2.)

    assert len(nest.GetConnections()) == net.edge_nb() + 3*net.node_nb()

    ns.set_step_currents(gids[::2], times=[40., 60.], currents=[800., 0.])

    min_conn = net.edge_nb() + 3*net.node_nb() + 1
    max_conn = net.edge_nb() + 4*net.node_nb()

    num_conn = len(nest.GetConnections())

    assert min_conn <= num_conn <= max_conn

    # check randomization of neuronal properties
    vms = {d['V_m'] for d in nest.GetStatus(gids)}

    assert len(vms) == 1

    ns.randomize_neural_states(net, {'V_m': ('uniform', -70., -60.)})

    vms = {d['V_m'] for d in nest.GetStatus(gids)}

    assert len(vms) == net.node_nb()

    # monitoring nodes
    sd, _ = ns.monitor_groups(net.population, net)

    assert len(nest.GetConnections()) == num_conn + net.node_nb()

    vm, rec = ns.monitor_nodes(gids[0], nest_recorder='voltmeter',
                               params={'interval': resol})

    assert len(nest.GetConnections()) == num_conn + net.node_nb() + 1

    nest.Simulate(100.)

    ns.plot_activity(show=False)

    ns.plot_activity(vm, rec, show=True)


if __name__ == "__main__":
    if not nngt.get_config("mpi"):
        test_net_creation()
        test_utils()



#cython: language_level=3
#!/usr/bin/env cython
#-*- coding:utf-8 -*-

""" Cython headers for C++ parallel generation tools for NNGT """

from libcpp.vector cimport vector
from libcpp.string cimport string
from libcpp cimport bool

from libc.stdint cimport int64_t, uint64_t


# ---------------------- #
# Load the c++ functions #
# ---------------------- #

cdef extern from "func_connect.h" namespace "generation":
    cdef void _gen_edges(
      int64_t* ia_edges, const vector[size_t]& first_nodes,
      const vector[unsigned int]& degrees, const vector[size_t]& second_nodes,
      const vector[ vector[size_t] ]& existing_edges, unsigned int idx,
      bool multigraph, bool directed, vector[long]& seeds) except +

    cdef void _cdistance_rule(
      int64_t* ia_edges, const vector[size_t]& source_nodes,
      const vector[ vector[size_t] ]& target_nodes, const string& rule,
      float scale, float norm, const vector[float]& x, const vector[float]& y,
      size_t num_neurons, size_t num_edges,
      const vector[ vector[size_t] ]& existing_edges, vector[float]& dist,
      bool multigraph, bool directed, vector[long]& seeds) except +



#-*- coding:utf-8 -*-
#
# This file is part of the NNGT project to generate and analyze
# neuronal networks and their activity.
# Copyright (C) 2015-2019  Tanguy Fardet
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

''' Node and edge attributes '''

import numpy as np

import nngt
import nngt.generation as ng


''' -------------- #
# Generate a graph #
# -------------- '''

# create an empty graph
g = nngt.Graph()

# add some nodes
g.new_node(10)            # create nodes 0, 1, ... to 9
print(g.node_nb(), '\n')  # returns 10

# add edges
g.new_edge(1, 4)          # create one connection going from 1 to 4
print(g.edge_nb())        # returns 1
g.new_edges([(0, 3), (5, 9), (9, 3)])
print(g.edge_nb(), '\n')  # returns 4


''' --------------- #
# Adding attributes #
# --------------- '''

g2 = nngt.Graph()

# add a new node with attributes
attributes = {
    'size': 2.,
    'color': 'blue',
    'a': 5,
    'blob': []
}

attribute_types = {
    'size': 'double',
    'color': 'string',
    'a': 'int',
    'blob': 'object'
}

g2.new_node(attributes=attributes, value_types=attribute_types)
print(g2.node_attributes, '\n')

# by default all nodes will have these properties with "empty" values
g2.new_node(2)
# for a double attribute like 'size', default value is NaN
print(g2.get_node_attributes(name="size"))
# for a string attribute like 'color', default value is ""
print(g2.get_node_attributes(name="color"))
# for an int attribute like 'a', default value is 0
print(g2.get_node_attributes(name='a'))
# for an object attribute like 'blob', default value is None
print(g2.get_node_attributes(name='blob'), '\n')

# attributes for multiple nodes can be set simultaneously
g2.new_node(3, attributes={'size': [4., 5., 1.], 'color': ['r', 'g', 'b']},
            value_types={'size': 'double', 'color': 'string'})
print(g2.node_attributes['size'])
print(g2.node_attributes['color'], '\n')

# creating attributes afterwards
import numpy as np
g3 = nngt.Graph(nodes=100)
g3.new_node_attribute('size', 'double',
                      values=np.random.uniform(0, 20, 100))
print(g3.node_attributes['size'][:5], '\n')

# edges attributes
edges = g3.new_edges(np.random.randint(0, 50, (10, 2)), ignore_invalid=True)
g3.new_edge_attribute('rank', 'int')
g3.set_edge_attribute('rank', val=2, edges=edges[:3, :])
print(g3.edge_attributes['rank'], '\n')

# check default values
e = g3.new_edge(99, 98)
g3.new_edges(np.random.randint(50, 100, (5, 2)), ignore_invalid=True)
print(g3.edge_attributes['rank'], '\n')


''' ---------------------------------------- #
# Generate and analyze more complex networks #
# ---------------------------------------- '''

from nngt import generation as ng
from nngt import analysis as na
from nngt import plot as nplt

# make an ER network
g = ng.erdos_renyi(nodes=1000, avg_deg=100)

if nngt.get_config("with_plot"):
    nplt.degree_distribution(g, ('in', 'total'), show=False)

print("Clustering ER: {}".format(na.global_clustering(g)))

# then a scale-free network
g = ng.random_scale_free(1.8, 3.2, nodes=1000, avg_deg=100)

if nngt.get_config("with_plot"):
    nplt.degree_distribution(g, ('in', 'out'), num_bins=30, logx=True,
                             logy=True, show=True)

print("Clustering SF: {}".format(na.global_clustering(g)))



#-*- coding:utf-8 -*-
#
# lib/nngt_config.py
#
# This file is part of the NNGT project, a graph-library for standardized and
# and reproducible graph analysis: generate and analyze networks with your
# favorite graph library (graph-tool/igraph/networkx) on any platform, without
# any change to your code.
# Copyright (C) 2015-2021 Tanguy Fardet
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

""" Configuration tools for NNGT """

import os
import sys
import logging
import importlib.util as imputil
from copy import deepcopy
from importlib import reload

import numpy as np

import nngt
from .errors import InvalidArgument
from .logger import _configure_logger, _init_logger, _log_message
from .rng_tools import seed as nngt_seed
from .test_functions import mpi_checker, num_mpi_processes, mpi_barrier


logger = logging.getLogger(__name__)


# ----------------- #
# Getter and setter #
# ----------------- #

def get_config(key=None, detailed=False):
    '''
    Get the NNGT configuration as a dictionary.

    Note
    ----
    This function has no MPI barrier on it.
    '''
    if key is None:
        cfg = {
            key: val.copy() if isinstance(val, list) else val
            for key, val in nngt._config.items()
        }

        if detailed:
            return cfg
        else:
            # hide mpi conf if not used
            if not nngt._config["mpi"]:
                del cfg['mpi_comm']

            # hide technical stuff
            del cfg["graph"]
            del cfg["library"]
            del cfg["palette_continuous"]
            del cfg["palette_discrete"]
            del cfg["use_tex"]
            del cfg["mpl_backend"]
            del cfg["color_lib"]

            # hide database config if not used
            rm = []
            if not nngt._config["use_database"]:
                for k in cfg:
                    if k.startswith('db_'):
                        rm.append(k)

            # hide log config
            for k in cfg:
                if k.startswith('log_'):
                    rm.append(k)

            for k in rm:
                del cfg[k]

        return cfg

    res = nngt._config[key]

    return res


@mpi_barrier
def set_config(config, value=None, silent=False):
    '''
    Set NNGT's configuration.

    Parameters
    ----------
    config : dict or str
        Either a full configuration dictionary or one key to be set together
        with its associated value.
    value : object, optional (default: None)
        Value associated to `config` if `config` is a key.

    Examples
    --------

    >>> nngt.set_config({'multithreading': True, 'omp': 4})
    >>> nngt.set_config('multithreading', False)

    Notes
    -----
    See the config file `nngt/nngt.conf.default` or `~/.nngt/nngt.conf` for
    details about your configuration.

    This function has an MPI barrier on it, so it must always be called on all
    processes.

    See also
    --------
    :func:`~nngt.get_config`
    '''
    old_mt     = nngt._config["multithreading"]
    old_mpi    = nngt._config["mpi"]
    old_omp    = nngt._config["omp"]
    old_gl     = nngt._config["backend"]
    old_msd    = nngt._config["msd"]

    old_config = nngt._config.copy()
    new_config = None

    if not isinstance(config, dict):
        new_config = {config: value}
    else:
        new_config = config.copy()

    for key, val in new_config.items():
        # support for previous "palette" keyword
        if key not in nngt._config:
            if key == "palette":
                _log_message(logger, "WARNING",
                             "`palette` argument is deprecated and will be "
                             "removed in version 3.")
            else:
                raise KeyError(
                    "Unknown configuration property: {}".format(key))

        if key == "log_level":
            new_config[key] = _convert(val)
        if key == "backend" and val != old_gl:
            nngt.use_backend(val)
        if key == "log_folder":
            new_config["log_folder"] = os.path.abspath(
                os.path.expanduser(val))
        if key == "db_folder":
            new_config["db_folder"] = os.path.abspath(
                os.path.expanduser(val))

    # support for previous "palette" keyword
    if "palette" in new_config:
        new_config["palette_continuous"] = new_config["palette"]
        new_config["palette_discrete"]   = new_config["palette"]

        del new_config["palette"]

    # check multithreading status and number of threads
    _pre_update_parallelism(new_config, old_mt, old_omp, old_mpi)

    # update
    nngt._config.update(new_config)

    # apply multithreading parameters
    _post_update_parallelism(new_config, old_gl, old_msd, old_mt, old_mpi)

    # update matplotlib
    if nngt._config['use_tex']:
        import matplotlib
        matplotlib.rc('text', usetex=True)

    # update database
    if nngt._config["use_database"] and not hasattr(nngt, "db"):
        from .. import database
        sys.modules["nngt.database"] = database
        if nngt._config["db_to_file"]:
            _log_message(logger, "WARNING",
                        "This functionality is not available")

    # update nest
    if imputil.find_spec("nest") is not None:
        _lazy_load("nngt.simulation")
        nngt._config["with_nest"] = True
    else:
        nngt._config["with_nest"] = False

    # check geometry
    try:
        import svg.path
        has_svg = True
    except:
        has_svg = False
    try:
        import dxfgrabber
        has_dxf = True
    except:
        has_dxf = False
    try:
        import shapely
        has_shapely = shapely.__version__
    except:
        has_shapely = False

    # check geospatial
    has_geospatial = False

    if imputil.find_spec("geopandas") is not None and has_shapely:
        _lazy_load("nngt.geospatial")
        has_geospatial = True

    # log changes
    _configure_logger(nngt._logger)

    glib = (nngt._config["library"] if nngt._config["library"] is not None
            else nngt)

    num_mpi = num_mpi_processes()

    s_mpi = False if not nngt._config["mpi"] \
            else "True ({} process{})".format(
                num_mpi, "es" if num_mpi > 1 else "")

    conf_info = _config_info.format(
        gl      = nngt._config["backend"] + " " + glib.__version__[:5],
        thread  = nngt._config["multithreading"],
        plot    = nngt._config["with_plot"],
        nest    = nngt._config["with_nest"],
        db      = nngt._config["use_database"],
        omp     = nngt._config["omp"],
        s       = "s" if nngt._config["omp"] > 1 else "",
        mpi     = s_mpi,
        shapely = has_shapely,
        svg     = has_svg,
        dxf     = has_dxf,
        geotool = has_geospatial,
    )

    if not silent and old_config != nngt._config:
        _log_conf_changed(conf_info)


# ----- #
# Tools #
# ----- #

def _convert(value):
    value = str(value)
    if value.isdigit():
        return int(value)
    elif value.lower() == "true":
        return True
    elif value.lower() == "false":
        return False
    elif value.upper() == "CRITICAL":
        return logging.CRTICAL
    elif value.upper() == "DEBUG":
        return logging.DEBUG
    elif value.upper() == "ERROR":
        return logging.ERROR
    elif value.upper() == "INFO":
        return logging.INFO
    elif value.upper() == "WARNING":
        return logging.WARNING
    else:
        return value


def _load_config(path_config):
    ''' Load `~/.nngt.conf` and parse it, return the settings '''
    with open(path_config, 'r') as fconfig:
        options = [l.strip() for l in fconfig if l.strip() and l[0] != "#"]
        for opt in options:
            sep = opt.find("=")
            opt_name = opt[:sep].strip()
            nngt._config[opt_name] = _convert(opt[sep+1:].strip())
    _init_logger(nngt._logger)


@mpi_checker(logging=True)
def _log_conf_changed(conf_info):
    logger.info(conf_info)


def _set_gt_config(old_gl, new_config):
    using_gt  = old_gl == "graph-tool"
    using_gt *= new_config.get("backend", old_gl) == "graph-tool"
    using_gt *= nngt._config["library"] is not None

    if "omp" in new_config and using_gt:
        omp_nest = new_config["omp"]
        if nngt._config['with_nest']:
            import nest
            omp_nest = nest.GetKernelStatus("local_num_threads")
        if omp_nest == new_config["omp"]:
            nngt._config["library"].openmp_set_num_threads(nngt._config["omp"])
        else:
            _log_message(logger, "WARNING",
                         "Using NEST and graph_tool, OpenMP number must be "
                         "consistent throughout the code. Current NEST "
                         "config states omp = " + str(omp_nest) + ", hence "
                         "`graph_tool` configuration was not changed.")


def _pre_update_parallelism(new_config, old_mt, old_omp, old_mpi):
    mt = "multithreading"

    if "omp" in new_config:
        if new_config["omp"] > 1:
            if mt in new_config and not new_config[mt]:
                _log_message(logger, "WARNING",
                             "Updating to 'multithreading' == False with "
                             "'omp' greater than one.")
            elif mt not in new_config and not old_mt:
                new_config[mt] = True
                _log_message(logger, "INFO",
                             "'multithreading' was set to False but new "
                             "'omp' is greater than one. Updating "
                             "'multithreading' to True.")

    if new_config.get('mpi', False) and new_config.get(mt, False):
        raise InvalidArgument('Cannot set both "mpi" and "multithreading" to '
                              'True simultaneously, choose one or the other.')
    elif new_config.get(mt, False):
        new_config['mpi'] = False
    elif new_config.get('mpi', False):
        if old_mt:
            new_config[mt] = False
            _log_message(logger, "INFO",
                         '"mpi" set to True but previous configuration was '
                         'using OpenMP; setting "multithreading" to False '
                         'to switch to mpi algorithms.')

    with_mt  = new_config.get(mt, old_mt)
    with_mpi = new_config.get('mpi', old_mpi)

    # check that seeds are correct
    if new_config.get('seeds', None) is not None:
        seeds = new_config['seeds']
        err  = 'Expected {} seeds.'
        err2 = 'All seeds must be different.'
        if with_mpi:
            from mpi4py import MPI
            comm = MPI.COMM_WORLD
            size = comm.Get_size()
            assert size == len(seeds), err.format(size)
            assert len(set(seeds)) == len(seeds), err2
        elif with_mt:
            num_omp = new_config.get("omp", old_omp)
            assert num_omp == len(seeds), err.format(num_omp)
            assert len(set(seeds)) == len(seeds), err2
    else:
        # reset seeds if necessary
        # - because the number of threads changed
        reset_seeds  = (new_config.get("omp", 1) != nngt._config["omp"])
        # - because we switched from OpenMP to MPI
        reset_seeds += (with_mpi and old_mt)
        # - because we switched from MPI to OpenMP
        reset_seeds += (with_mt and old_mpi)

        if reset_seeds:
            new_config['seeds'] = None
            new_config['msd']   = None
            nngt._seeded        = False


def _post_update_parallelism(new_config, old_gl, old_msd, old_mt, old_mpi):
    # reload for omp
    new_multithreading = new_config.get("multithreading", old_mt)

    if new_multithreading != old_mt:
        reload(sys.modules["nngt"].generation.graph_connectivity)
        reload(sys.modules["nngt"].generation.connectors)
        reload(sys.modules["nngt"].generation.rewiring)

    # if multithreading loading failed, set omp back to 1
    if not nngt._config['multithreading']:
        nngt._config['omp'] = 1
        nngt._config['seeds'] = None

    # if MPI is on, set mpi_comm and check random numbers
    if new_config.get('mpi', old_mpi):
        from mpi4py import MPI
        comm = MPI.COMM_WORLD
        rank = comm.Get_rank()
        nngt._config['mpi_comm'] = comm
        # check that master seed is the same everywhere
        msd = nngt._config['msd']
        msd = comm.gather(msd, root=0)
        if rank == 0:
            if None not in msd:
                msd = np.array(msd, dtype=int)
                if not np.alltrue(msd == msd[0]):
                    nngt._config["mpi"] = False
                    raise InvalidArgument("'msd' entry must be the same on "
                                          "all MPI processes.")
            else:
                differs = [seed != None for seed in msd]
                if np.any(differs):
                    raise InvalidArgument("'msd' entry must be the same on "
                                          "all MPI processes.")

    # reload for mpi
    if new_config.get('mpi', old_mpi) != old_mpi:
        reload(sys.modules["nngt"].generation.graph_connectivity)
        reload(sys.modules["nngt"].generation.connectors)
        reload(sys.modules["nngt"].generation.rewiring)

    # set graph-tool config
    _set_gt_config(old_gl, new_config)

    # seed python RNGs
    if old_msd != nngt._config['msd'] or not nngt._seeded:
        nngt_seed(msd=nngt._config['msd'])


def _lazy_load(fullname):
    '''
    Lazy loading for simulation.

    From: https://stackoverflow.com/a/51126745/5962321
    '''
    try:
        return sys.modules[fullname]
    except KeyError:
        spec   = imputil.find_spec(fullname)
        module = imputil.module_from_spec(spec)
        loader = imputil.LazyLoader(spec.loader)

        # setup module and insert into sys.modules
        loader.exec_module(module)

        return module


_config_info = '''
# ----------- #
# NNGT config #
# ----------- #
Graph library:  {gl}
Multithreading: {thread} ({omp} thread{s})
MPI:            {mpi}
Plotting:       {plot}
NEST support:   {nest}
Shapely:        {shapely}
SVG support:    {svg}
DXF support:    {dxf}
Database:       {db}
Geo support:    {geotool}
'''



#-*- coding:utf-8 -*-
#
# database/db_main.py
#
# This file is part of the NNGT project, a graph-library for standardized and
# and reproducible graph analysis: generate and analyze networks with your
# favorite graph library (graph-tool/igraph/networkx) on any platform, without
# any change to your code.
# Copyright (C) 2015-2021 Tanguy Fardet
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

""" Main object for database management """

import platform
from datetime import datetime
from itertools import permutations

try:
    import nest
except ImportError:
    raise ImportError("Database module requires NEST to work.")

import peewee

import nngt
from nngt.lib.db_tools import psutil
from .db_generation import *


__all__ = ['NNGTdb']


class NNGTdb:
    '''
    Class containing the database object and methods to store the simulation
    properties.
    '''
    
    tables = {
        'activity': Activity,
        'computer': Computer,
        'connection': Connection,
        'neuralnetwork': NeuralNetwork,
        'neuron': Neuron,
        'simulation': Simulation,
        'synapse': Synapse
    }

    def __init__(self):
        self.db = nngt._main_db
        self.db.connect()
        self.db.create_tables(self.tables.values(), safe=True)
        self._update_models()
        self.activity = None
        self.current_simulation = None
        self.simulator_lib = None
        self.computer = None
        self.neuralnet = None
        self.connections = {}
        self.nodes = {}
    
    def _update_class(self, table, **kwargs):
        ''' Add a field for each property of the considered node. '''
        klass = self.tables[table]
        columns = [ x.name for x in self.db.get_columns(table) ]
        migrator = db_migrator[self.db.__class__.__name__](self.db)
        type_names = False
        if "dtype" in kwargs:
            type_names = kwargs["dtype"]
            del kwargs["dtype"]
        for attr, value in kwargs.items():
            if attr not in ignore:
                # generate field instance
                dtype = value if type_names else value.__class__.__name__
                val_field = val_to_field[dtype](null=True)
                if len(attr) == 1 and attr.isupper():
                    attr = 2*attr.lower()
                klass._meta.add_field(attr, val_field)
                # check whether the column exists on the table, if not create
                if not attr in columns:
                    try:
                        migrate(migrator.add_column(table, attr, val_field))
                    except peewee.OperationalError:
                        pass
        return klass
    
    def _update_models(self):
        ''' Update the models so that we can query the database with them '''
        tables = self.db.get_tables()
        for table in tables:
            delete = []
            col_names = self.db.get_columns(table)
            col_dict = { x.name: x.data_type for x in col_names}
            for attr in iter(col_dict.keys()):
                if attr == "compressed_file":
                    col_dict["compressed_file"] = "compressed"
                elif "_id" in attr:
                    delete.append(attr)
            col_dict["dtype"] = True
            for key in delete:
                del col_dict[key]
            self.tables[table] = self._update_class(table, **col_dict)

    def _make_computer_entry(self):
        ''' Get the computer properties.

        Returns
        -------
        computer : :class:`~nngt.database.Computer`
            New computer entry.
        '''
        name = platform.node()
        comp_prop = {
            'name': name if name else "unknown",
            'platform': platform.platform(),
            'python': platform.python_version(),
            'cores': psutil.cpu_count(),
            'ram': psutil.virtual_memory().total,
        }
        computer = Computer(**comp_prop)
        return computer

    def _make_network_entry(self, network):
        '''
        Get the network properties.

        Parameters
        ----------
        network : :class:`~nngt.Network` or subclass
            Network used in the current simulation.

        Returns
        -------
        neuralnet : :class:`~nngt.database.NeuralNetwork`
            New NeuralNetwork entry.
        '''
        if network is not None:
            weighted = network.is_weighted()
            net_prop = {
                'network_type': network.type,
                'directed': network.is_directed(),
                'nodes': network.node_nb(),
                'edges': network.edge_nb(),
                'weighted': weighted,
                'compressed_file': str(network).encode('utf-8')
            }
            if weighted:
                net_prop['weight_distribution'] = network._w
            neuralnet = NeuralNetwork(**net_prop)
        else:
            neuralnet = NeuralNetwork()
        return neuralnet

    def _make_neuron_entry(self, network, group):
        '''
        Get the neuronal properties.

        Parameters
        ----------
        network : :class:`~nngt.Network` or subclass
            Network used in the current simulation.
        group : :class:`~nngt.core.NeuralGroup`
            Group which properties will be fetched.

        Returns
        -------
        neuron : :class:`~nngt.database.Neuron`
            New neuron entry.
        '''
        nngt_id = group.ids[0]
        gid = network.nest_gids[nngt_id]
        # get the dictionary
        neuron_prop = nest.GetStatus((gid,))[0]
        # update Neuron class accordingly
        Neuron = self._update_class("neuron", **neuron_prop)
        neuron = Neuron(**neuron_prop)
        return neuron

    def _make_synapse_entry(self, network, group_pre, group_post):
        '''
        Get the synaptic properties.

        Parameters
        ----------
        network : :class:`~nngt.Network` or subclass
            Network used in the current simulation.
        group_pre : :class:`~nngt.core.NeuralGroup`
            Pre-synaptic group.
        group_post : :class:`~nngt.core.NeuralGroup`
            Post-synaptic group.

        Returns
        -------
        synapse : :class:`~nngt.database.Synapse`
            New synapse entry.
        '''
        syn_model = "static_synapse"
        if (group_pre.name, group_post.name) in network.population.syn_spec:
            pre, post = group_pre.name, group_post.name
            syn_model = network.population.syn_spec[(pre, post)]
            if isinstance(syn_model, dict):
                syn_model = syn_model.get("model", "static_synapse")
        source_gids = tuple(network.nest_gids[group_pre.ids])
        target_gids = tuple(network.nest_gids[group_post.ids])
        connections = nest.GetConnections(
            synapse_model=syn_model, source=source_gids, target=target_gids)
        # get the dictionary
        syn_prop={}
        if connections:
            syn_prop = nest.GetStatus((connections[0],))[0]
            # check single uppercase letters
            sngl_upper = []
            for k in syn_prop:
                if len(k) == 1 and k.isupper():
                    sngl_upper.append(k)
            for k in sngl_upper:
                syn_prop[2*k.lower()] = syn_prop[k]
                del syn_prop[k]
            # update Synapse class accordingly
            Synapse = self._update_class("synapse", **syn_prop)
        synapse = Synapse(**syn_prop)
        return synapse

    def _make_connection_entry(self, neuron_pre, neuron_post, synapse):
        '''
        Create the entries for the Connections table from a list of
        (pre, post, syn) triples.

        Parameters
        ----------
        neuron_pre : :class:`~nngt.database.Neuron`
            Pre-synaptic neuron entry.
        neuron_post : :class:`~nngt.database.Neuron`
            Post-synaptic neuron entry.
        synapse : :class:`~nngt.database.Synapse`
            Synapse entry.
        
        Returns
        -------
        :class:`~nngt.database.Connection`
        '''
        return Connection(pre=neuron_pre, post=neuron_post, synapse=synapse)

    def _get_simulation_prop(self, network, simulator):
        '''
        Get the simulation properties.
        
        Parameters
        ----------
        network : :class:`~nngt.Network`
            Network used for the simulation.
        simulator : str
            Name of the simulator use (NEST, BRIAN...).

        Returns
        -------
        sim_prop : dict
            Dictionary containing the relevant key/value pairs to fill the
            :class:`~nngt.database.Simulation` class.
        '''
        pop, size = [], []
        for name, group in iter(network.population.items()):
            pop.append(name)
            size.append(len(group.ids))
        self.current_simulation = {
            'start_time': datetime.now(),
            'simulated_time': nest.GetKernelStatus('time'),
            'resolution': nest.GetKernelStatus('resolution'),
            'simulator': simulator.lower(),
            'grnd_seed': nest.GetKernelStatus('grng_seed'),
            'local_seeds': nest.GetKernelStatus('rng_seeds'),
            'population': pop,
            'pop_sizes': size
        }

    def _make_activity_entry(self, network=None):
        '''
        Create an activity entry from an
        :class:`~nngt.simulation.ActivityRecord` object.
        '''
        raster      = nngt.analysis.get_spikes(astype="np")
        activity    = nngt.simulation.analyze_raster(raster, network=network)
        di_activity = activity.properties
        di_activity["raster"] = raster
        act_attr = { k: v.__class__.__name__ for k, v in  di_activity.items() }
        if "spike_files" in act_attr:
            act_attr["spike_files"] = "compressed"
        act_attr["dtypes"] = True
        ''' ..todo ::
            compress the spike files '''
        Activity = self._update_class("activity", **act_attr)
        activity_entry = Activity(**di_activity)
        self.current_simulation['activity'] = activity_entry
        return activity_entry
        
    def log_simulation_start(self, network, simulator, save_network=True):
        '''
        Record the simulation start time, all nodes, connections, network, and
        computer properties, as well as some of simulation's.

        Parameters
        ----------
        network : :class:`~nngt.Network` or subclass
            Network used for the current simulation.
        simulator : str
            Name of the simulator.
        save_network : bool, optional (default: True)
            Whether to save the network or not.
        '''
        if not self.is_clear():
            raise RuntimeError("Database log started without clearing the "
                               "previous one.")
        self._get_simulation_prop(network, simulator)
        # computer and network data
        self.computer  = self._make_computer_entry()
        self.neuralnet = (self._make_network_entry(network)
                          if save_network else None)
        self.current_simulation['computer'] = self.computer
        self.current_simulation['network']  = self.neuralnet
        # neurons, synapses and connections
        perm_names = tuple(permutations(network.population.keys(), 2))
        perm_groups = tuple(permutations(network.population.values(), 2))
        if not perm_names:
            group_name = list(network.population.keys())[0]
            group = network.population[group_name]
            perm_names = ((group_name, group_name),)
            perm_groups = ((group, group),)
        self.nodes = {}
        for (name_pre, name_post), (pre, post) in zip(perm_names, perm_groups):
            if name_pre not in self.nodes:
                self.nodes[name_pre] = self._make_neuron_entry(network, pre)
            if name_post not in self.nodes:
                self.nodes[name_post] = self._make_neuron_entry(network, post)
            synapse = self._make_synapse_entry(network, pre, post)
            self.nodes["syn_{}->{}".format(name_pre, name_post)] = synapse
            conn = self._make_connection_entry(self.nodes[name_pre],
                                               self.nodes[name_post], synapse)
            self.connections["{}->{}".format(name_pre, name_post)] = conn
        

    def log_simulation_end(self, network=None, log_activity=True):
        '''
        Record the simulation completion and simulated times, save the data,
        then reset.
        '''
        if self.is_clear():
            raise RuntimeError("Database log ended with empy log.")
        # get completion time and simulated time
        self.current_simulation['completion_time'] = datetime.now()
        start_time = self.current_simulation['simulated_time']
        new_time   = nest.GetKernelStatus('time')
        self.current_simulation['simulated_time'] = new_time - start_time
        # save activity if provided
        if log_activity:
            self.activity = self._make_activity_entry(network)
        else:
            self.activity = Activity()
            self.current_simulation['activity'] = self.activity
        # save data and reset
        self.activity.save()
        self.computer.save()
        if self.neuralnet is not None:
            self.neuralnet.save()
        for entry in iter(self.nodes.values()):
            entry.save()
        for entry in iter(self.connections.values()):
            entry.save()
        simul_data = Simulation(**self.current_simulation)
        simul_data.save()
        # ~ if nngt.get_config("db_to_file"):
            # ~ from .csv_utils import dump_csv
            # ~ db_cls = list(self.tables.values())
            # ~ q = (Simulation.select(*db_cls).join(Computer).switch(Simulation)
                  # ~ .join(NeuralNetwork).switch(Simulation).join(Activity)
                  # ~ .switch(Simulation).join(Connection).join(Neuron, on=Connection.pre)
                  # ~ .switch(Connection).join(Neuron, on=Connection.post)
                  # ~ .switch(Connection).join(Synapse)).select(*db_cls)
            # ~ dump_csv(q, "{}_{}.csv".format(self.computer.name,
                                           # ~ simul_data.completion_time))
        self.reset()
    
    def get_results(self, table, column=None, value=None):
        '''
        Return the entries where the attribute `column` satisfies the required
        equality.
        
        Parameters
        ----------
        table : str
            Name of the table where the search should be performed (among
            ``'simulation'``, ``'computer'``, ``'neuralnetwork'``,
            ``'activity'``, ``'synapse'``, ``'neuron'``, or ``'connection'``).
        column : str, optional (default: None)
            Name of the variable of interest (a column on the table). If None,
            the whole table is returned.
        value : `column` corresponding type, optional (default: None)
            Specific value for the variable of interest. If None, the whole
            column is returned.
        
        Returns
        -------
        :class:`peewee.SelectQuery` with entries matching the request.
        '''
        TableModel = self.tables[table]
        if column is None:
            return TableModel.select()
        elif value is None:
            return TableModel.select(getattr(TableModel, column))
        else:
            return TableModel.select().where(
                getattr(TableModel, column) == value)

    def is_clear(self):
        ''' Check that the logs are clear. '''
        clear = True
        clear *= self.current_simulation is None
        clear *= self.simulator_lib is None
        clear *= self.computer is None
        clear *= self.neuralnet is None
        clear *= not self.nodes
        clear *= not self.connections
        return clear

    def reset(self):
        ''' Reset log status. '''
        self.current_simulation = None
        self.simulator_lib = None
        self.computer = None
        self.neuralnet = None
        self.connections = {}
        self.nodes = {}




#-*- coding:utf-8 -*-
#
# This file is part of the NNGT project to generate and analyze
# neuronal networks and their activity.
# Copyright (C) 2015-2020  Tanguy Fardet
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
Plot various graph properties
=============================
"""

import nngt
import nngt.plot as nplt
from nngt.geometry import Shape

import matplotlib.pyplot as plt

plt.rcParams.update({
    'axes.edgecolor': 'grey', 'xtick.color': 'grey', 'ytick.color': 'grey',
    "figure.facecolor": (0, 0, 0, 0), "axes.facecolor": (0, 0, 0, 0),
    "axes.labelcolor": "grey", "axes.titlecolor": "grey", "text.color": "grey"
})


nngt.seed(0)


# %%
# Let's start by making a random exponential graph

shape = Shape.disk(100)

g = nngt.generation.distance_rule(5, shape=shape, nodes=1000, avg_deg=20)


# %%
# Let's plot the distances

nplt.edge_attributes_distribution(g, "distance", show=True)

# %%
# We then compute the betweenness and see how it correlates with the distance

nbetw, ebetw = nngt.analysis.betweenness(g)

g.new_edge_attribute("betweenness", "float", values=ebetw)

nplt.correlation_to_attribute(g, "distance", "betweenness",
                              attribute_type="edge", show=True)


# %%
# Let's check the correlations between various node properties and their degree

g.new_node_attribute("betweenness", "float", values=nbetw)

attr = ["betweenness", "clustering", "in-degree", "subgraph_centrality"]

nplt.correlation_to_attribute(g, "out-degree", attr, show=True)



#-*- coding:utf-8 -*-
#
# core/spatial_graph.py
#
# This file is part of the NNGT project, a graph-library for standardized and
# and reproducible graph analysis: generate and analyze networks with your
# favorite graph library (graph-tool/igraph/networkx) on any platform, without
# any change to your code.
# Copyright (C) 2015-2021 Tanguy Fardet
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

""" SpatialGraph class for spatial graph generation and management """

import numpy as np

import nngt
from nngt.lib import InvalidArgument, nonstring_container

from .connections import Connections
from .graph import Graph


# ------------ #
# SpatialGraph #
# ------------ #

class SpatialGraph(Graph):

    """
    The detailed class that inherits from :class:`~nngt.Graph` and implements
    additional properties to describe spatial graphs (i.e. graph where the
    structure is embedded in space.
    """

    #-------------------------------------------------------------------------#
    # Class properties

    __num_graphs = 0
    __max_id = 0

    #-------------------------------------------------------------------------#
    # Constructor, destructor, attributes

    def __init__(self, nodes=0, name="SpatialGraph", weighted=True,
                 directed=True, copy_graph=None, shape=None, positions=None,
                 **kwargs):
        '''
        Initialize SpatialClass instance.

        .. versionchanged: 2.4
            Move `from_graph` to `copy_graph` to reflect changes in Graph.

        Parameters
        ----------
        nodes : int, optional (default: 0)
            Number of nodes in the graph.
        name : string, optional (default: "Graph")
            The name of this :class:`Graph` instance.
        weighted : bool, optional (default: True)
            Whether the graph edges have weight properties.
        directed : bool, optional (default: True)
            Whether the graph is directed or undirected.
        shape : :class:`~nngt.geometry.Shape`, optional (default: None)
            Shape of the neurons' environment (None leads to a square of
            side 1 cm)
        positions : :class:`numpy.array` (N, 2), optional (default: None)
            Positions of the neurons; if not specified and `nodes` is not 0,
            then neurons will be reparted at random inside the
            :class:`~nngt.geometry.Shape` object of the instance.
        **kwargs : keyword arguments for :class:`~nngt.Graph` or
            :class:`~nngt.geometry.Shape` if no shape was given.

        Returns
        -------
        self : :class:`~nggt.SpatialGraph`
        '''
        self.__id = self.__class__.__max_id
        self.__class__.__num_graphs += 1
        self.__class__.__max_id += 1

        self._shape = None
        self._pos   = None

        super().__init__(nodes, name=name, weighted=weighted,
                         directed=directed, copy_graph=copy_graph, **kwargs)

        self._init_spatial_properties(shape, positions, **kwargs)

    def __del__(self):
        if hasattr(self, '_shape'):
            if self._shape is not None:
                self._shape._parent = None
            self._shape = None

        super().__del__()

        self.__class__.__num_graphs -= 1

    @property
    def shape(self):
        ''' The environment's spatial structure. '''
        return self._shape

    #-------------------------------------------------------------------------#
    # Init tool

    def _init_spatial_properties(self, shape, positions=None, **kwargs):
        '''
        Create the positions of the neurons from the graph `shape` attribute
        and computes the connections distances.
        '''
        positions = None if positions is None else np.asarray(positions)

        self.new_edge_attribute('distance', 'double')

        if positions is not None and len(positions) != self.node_nb():
            raise InvalidArgument("Wrong number of neurons in `positions`.")

        if shape is not None:
            shape.set_parent(self)
            self._shape = shape
        else:
            if positions is None or not np.any(positions):
                if 'height' in kwargs and 'width' in kwargs:
                    self._shape = nngt.geometry.Shape.rectangle(
                        kwargs['height'], kwargs['width'], parent=self)
                elif 'radius' in kwargs:
                    self._shape = nngt.geometry.Shape.disk(
                        kwargs['radius'], parent=self)
                elif 'radii' in kwargs:
                    self._shape = nngt.geometry.Shape.ellipse(
                        kwargs['radii'], parent=self)
                elif 'polygon' in kwargs:
                    self._shape = nngt.geometry.Shape.from_polygon(
                        kwargs['polygon'], min_x=kwargs.get('min_x', -5000.),
                        max_x=kwargs.get('max_x', 5000.),
                        unit=kwargs.get('unit', 'um'), parent=self)
                else:
                    raise RuntimeError('SpatialGraph needs a `shape` or '
                                       'keywords arguments to build one, or '
                                       'at least `positions` so it can create '
                                       'a square containing them')
            else:
                minx, maxx = np.min(positions[:, 0]), np.max(positions[:, 0])
                miny, maxy = np.min(positions[:, 1]), np.max(positions[:, 1])

                height, width = 1.01*(maxy - miny), 1.01*(maxx - minx)

                centroid = (0.5*(maxx + minx), 0.5*(maxy + miny))

                self._shape = nngt.geometry.Shape.rectangle(
                    height, width, centroid=centroid, parent=self)

        b_rnd_pos = True if not self.node_nb() or positions is None else False
        self._pos = self._shape.seed_neurons() if b_rnd_pos else positions

        Connections.distances(self)

    #-------------------------------------------------------------------------#
    # Positions

    def get_positions(self, nodes=None):
        '''
        Returns a copy of the nodes' positions as a (N, 2) array.

        Parameters
        ----------
        nodes : int or array-like, optional (default: all nodes)
            List of the nodes for which the position should be returned.
        '''
        if nodes is not None:
            if nonstring_container(nodes):
                # numpy slicing does not work with everything
                nodes = np.asarray(nodes)

                return np.array(self._pos[nodes])
            else:
                return self._pos[nodes]

        return np.array(self._pos)

    def set_positions(self, positions, nodes=None):
        '''
        Set the nodes' positions as a (N, 2) array.

        Parameters
        ----------
        positions : array-like
            List of positions, of shape (N, 2).
        nodes : int or array-like, optional (default: all nodes)
            List of the nodes for which the position should be set.
        '''
        if nodes is not None:
            self._pos[nodes] = positions
        else:
            if len(positions) != self.node_nb():
                raise ValueError("One position per node is required.")
            self._pos = np.array(positions)



#!/usr/bin/env python
#-*- coding:utf-8 -*-
#
# This file is part of the NNGT project to generate and analyze
# neuronal networks and their activity.
# Copyright (C) 2015-2019  Tanguy Fardet
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

""" NNGT documentation build configuration file. """

import sys
import os, errno
import shlex
import re
import fnmatch
import importlib.util as imputil

import sphinx_bootstrap_theme


# Paths
current_directory = os.path.abspath('.')
sys.path.append(current_directory)
sys.path.append(current_directory + "/extensions")  # custom extensions folder

# Simlink to geometry.examples
src = os.path.abspath('../nngt/geometry/examples')
tgt = os.path.abspath('modules/examples/')

try:
    os.remove(tgt)               # remove existing simlink
except OSError as e:
    if e.errno != errno.ENOENT:  # errno.ENOENT = no such file or directory
        raise                    # raise if a different error occurred

os.symlink(src, tgt)


'''
If on rtd, the graph libraries are not available so they need to be mocked
'''

try:
    if imputil.find_spec("nest") is None:
        # NEST is not available
        raise ImportError
except:
    import mock
    mock_object = mock.Mock(__name__ = "Mock", __bases__ = (object,))

    class Mock(object):

        def __init__(self, *args, **kwargs):
            super(Mock, self).__init__()

        def __call__(self, *args, **kwargs):
            return self

        def __getattr__(self, name):
            return self

        def __getitem__(self, name):
            return self

        def __setitem__(self, name, value):
            pass

        @property
        def __version__(self):
            return ""

        def __iter__(self):
            return self

        def next(self):
            raise StopIteration

        def __next__(self):
            self.next()

        @property
        def __name__(self):
            return "Mock"

        @property
        def __bases__(self):
            return (object,)

    sys.modules["nest"] = Mock()


# -- Setup all autosum then start --------------------------------------------

# import nngt
import nngt

# import simulation & geospatial explicitely to avoid import conflict with
# lazy load
try:
    import nngt.geospatial
except:
    pass

import nngt.simulation

from autosum import gen_autosum

# find all *.in files

inputs = []
skip   = ('main-functions.rst.in', 'side-classes.rst.in', 'geometry.rst.in')

for root, dirnames, filenames in os.walk('.'):
    for filename in fnmatch.filter(filenames, '*.in'):
        if filename not in skip:
            inputs.append(os.path.join(root, filename))

# list of classes to ignore for each module
ignore = {
    'nngt.core': ("Graph", "Network", "SpatialGraph", "SpatialNetwork",
                  "Group", "MetaGroup", "MetaNeuralGroup", "NeuralPop",
                  "NeuralGroup"),
    'nngt.lib': ("custom", "decorate", "deprecated", "graph_tool_check",
                 "mpi_barrier", "mpi_checker", "mpi_random", "not_implemented",
                 "num_mpi_processes", "on_master_process", "seed"),
    'nngt.generation': ('connect_nodes', 'connect_groups', 'connect_types',
                        'connect_neural_types', 'random_rewire',
                        'lattice_rewire', 'connect_neural_groups')
}

for f in inputs:
    target = f[:-3]  # remove '.in'
    # find the module (what will come after nngt, it is the name of the file)
    last_dot = target.rfind('.')
    last_slash = target.rfind('/')
    module = target[last_slash + 1:last_dot]
    if module != 'nngt':
        module = 'nngt.' + module
    gen_autosum(f, target, module, 'full', ignore=ignore.get(module, None))

# Add nngt (functions)
source = current_directory + "/modules/nngt/main-functions.rst.in"
target = current_directory + "/modules/nngt/main-functions.rst"
gen_autosum(source, target, 'nngt', 'summary', dtype="func")
gen_autosum(target, target, 'nngt', 'autofunction', dtype="func")

# nngt (side classes)
source = current_directory + "/modules/nngt/side-classes.rst.in"
target = current_directory + "/modules/nngt/side-classes.rst"

gen_autosum(source, target, 'nngt', 'summary', dtype="class",
            ignore=("Graph", "Network", "SpatialGraph", "SpatialNetwork"))

gen_autosum(target, target, 'nngt', 'autoclass', dtype="class",
            ignore=("Graph", "Network", "SpatialGraph", "SpatialNetwork"))

# geometry
source = current_directory + "/modules/geometry.rst.in"
target = current_directory + "/modules/geometry.rst"
gen_autosum(source, target, 'nngt.geometry', 'summary', dtype="all")

# generation
source = current_directory + "/modules/generation.rst.in"
target = current_directory + "/modules/generation.rst"
gen_autosum(source, target, 'nngt.generation', 'summary', dtype="func",
            ignore=ignore['nngt.generation'])
gen_autosum(target, target, 'nngt.generation', 'autofunction', dtype="func")




# -- NNGT setup -----------------------------------------------------------

from nngt import __version__ as nngt_version

# set database
try:
    import peewee
    nngt.set_config("use_database", True)
except ImportError:
    pass


# -- General configuration ------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
#needs_sphinx = '1.0'

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    'sphinx.ext.imgmath',
    'sphinx.ext.viewcode',
    'sphinx.ext.intersphinx',
    'sphinx.ext.autodoc',
    'sphinx.ext.autosummary',
    'sphinx.ext.doctest',
    'sphinx.ext.napoleon',
    'linksourcecode',
    'extlinks_fancy',
    'sphinx_gallery.gen_gallery',
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:
# source_suffix = ['.rst', '.md']
source_suffix = '.rst'

# The encoding of source files.
#source_encoding = 'utf-8-sig'

# The master toctree document.
master_doc = 'index'

# General information about the project.
project = u'NNGT'
copyright = u'2015, Tanguy Fardet'
author = u'Tanguy Fardet'

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
#
# The short X.Y version.
version = nngt_version
while version[-1].isalpha():
    version = version[:-1]
while version.count('.') > 1:
    idx_last_dot = version.rfind('.')
    version = version[:idx_last_dot]
# The full version, including alpha/beta/rc tags.
release = nngt_version

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = None

# There are two options for replacing |today|: either, you set today to some
# non-false value, then it is used:
#today = ''
# Else, today_fmt is used as the format for a strftime call.
#today_fmt = '%B %d, %Y'

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
exclude_patterns = ['.build']

# The reST default role (used for this markup: `text`) to use for all
# documents.
#default_role = None

# If true, '()' will be appended to :func: etc. cross-reference text.
#add_function_parentheses = True

# If true, the current module name will be prepended to all description
# unit titles (such as .. function::).
#add_module_names = True

# If true, sectionauthor and moduleauthor directives will be shown in the
# output. They are ignored by default.
#show_authors = False

# A list of ignored prefixes for module index sorting.
#modindex_common_prefix = []

# If true, keep warnings as "system message" paragraphs in the built documents.
#keep_warnings = False

# If true, `todo` and `todoList` produce output, else they produce nothing.
todo_include_todos = False


# -- Options for HTML output ----------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#~ html_theme = 'sphinx_rtd_theme'

html_theme = 'nngt_theme'
html_theme_path = ["."] + sphinx_bootstrap_theme.get_html_theme_path()
html_use_smartypants = False

html_theme_options = {
    # A list of tuples containing pages or urls to link to.
    # Valid tuples should be in the following forms:
    #    (name, page)                 # a link to a page
    #    (name, "/aa/bb", 1)          # a link to an arbitrary relative url
    #    (name, "http://example.com", True) # arbitrary absolute url
    # Note the "1" or "True" value above as the third argument to indicate
    # an arbitrary url.
    'navbar_links': [
        ("Modules", "py-modindex"),
        ("Index", "genindex"),
        ("SourceHut", "https://git.sr.ht/~tfardet/NNGT", True),
        ("GitHub", "https://github.com/tfardet/NNGT", True),
    ],

    # Render the next and previous page links in navbar. (Default: true)
    'navbar_sidebarrel': False,

    # Render the current pages TOC in the navbar. (Default: true)
    'navbar_pagenav': True,

    # Tab name for the current pages TOC. (Default: "Page")
    'navbar_pagenav_name': "Current",

    # Global TOC depth for "site" navbar tab. (Default: 1)
    # Switching to -1 shows all levels.
    'globaltoc_depth': 2,

    # Include hidden TOCs in Site navbar?
    #
    # Note: If this is "false", you cannot have mixed ``:hidden:`` and
    # non-hidden ``toctree`` directives in the same page, or else the build
    # will break.
    #
    # Values: "true" (default) or "false"
    'globaltoc_includehidden': "true",

    # Fix navigation bar to top of page?
    # Values: "true" (default) or "false"
    'navbar_fixed_top': "false",

    # Location of link to source.
    # Options are "nav" (default), "footer" or anything else to exclude.
    'source_link_position': "",

    # Bootswatch (http://bootswatch.com/) theme.
    'bootswatch_theme': "yeti"
}

html_sidebars = {'**': ['customtoc.html', 'searchbox.html']}


# The name for this set of Sphinx documents.  If None, it defaults to
# "<project> v<release> documentation".
#html_title = None

# A shorter title for the navigation bar.  Default is the same as html_title.
#html_short_title = None

# The name of an image file (relative to this directory) to place at the top
# of the sidebar.
#~ html_logo = 'images/nngt_logo.png'
#~ html_logo = 'images/nngt_ico.png'

# The name of an image file (within the static path) to use as favicon of the
# docs.  This file should be a Windows icon file (.ico) being 16x16 or 32x32
# pixels large.
html_favicon = 'images/nngt_ico.png'

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
# Set differently depending on the location?

html_static_path = ['_static']

# Add permalinks to headers
html_permalinks_icon = "#"

# Add any extra paths that contain custom files (such as robots.txt or
# .htaccess) here, relative to this directory. These files are copied
# directly to the root of the documentation.
#html_extra_path = []

# If not '', a 'Last updated on:' timestamp is inserted at every page bottom,
# using the given strftime format.
#html_last_updated_fmt = '%b %d, %Y'

# If true, SmartyPants will be used to convert quotes and dashes to
# typographically correct entities.
#html_use_smartypants = True

# Custom sidebar templates, maps document names to template names.
#~ html_sidebars = {
   #~ '**': ['globaltoc.html', 'sourcelink.html', 'searchbox.html'],
#~ }

# Additional templates that should be rendered to pages, maps page names to
# template names.
#html_additional_pages = {}

# If false, no module index is generated.
#html_domain_indices = True

# If false, no index is generated.
#html_use_index = True

# If true, the index is split into individual pages for each letter.
#html_split_index = False

# If true, links to the reST sources are added to the pages.
#html_show_sourcelink = True

# If true, "Created using Sphinx" is shown in the HTML footer. Default is True.
#html_show_sphinx = True

# If true, "(C) Copyright ..." is shown in the HTML footer. Default is True.
#html_show_copyright = True

# If true, an OpenSearch description file will be output, and all pages will
# contain a <link> tag referring to it.  The value of this option must be the
# base URL from which the finished HTML is served.
#html_use_opensearch = ''

# This is the file name suffix for HTML files (e.g. ".xhtml").
#html_file_suffix = None

# Language to be used for generating the HTML full-text search index.
# Sphinx supports the following languages:
#   'da', 'de', 'en', 'es', 'fi', 'fr', 'hu', 'it', 'ja'
#   'nl', 'no', 'pt', 'ro', 'ru', 'sv', 'tr'
#html_search_language = 'en'

# A dictionary with options for the search language support, empty by default.
# Now only 'ja' uses this config value
#html_search_options = {'type': 'default'}

# The name of a javascript file (relative to the configuration directory) that
# implements a search results scorer. If empty, the default will be used.
#html_search_scorer = 'scorer.js'

# Output file base name for HTML help builder.
htmlhelp_basename = 'NNGTdoc'

# -- Options for LaTeX output ---------------------------------------------

latex_elements = {
# The paper size ('letterpaper' or 'a4paper').
#'papersize': 'letterpaper',

# The font size ('10pt', '11pt' or '12pt').
#'pointsize': '10pt',

# Additional stuff for the LaTeX preamble.
#'preamble': '',

# Latex figure (float) alignment
#'figure_align': 'htbp',
}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title,
#  author, documentclass [howto, manual, or own class]).
latex_documents = [
  (master_doc, 'NNGT.tex', u'NNGT Documentation',
   u'Tanguy Fardet', 'manual'),
]

# The name of an image file (relative to this directory) to place at the top of
# the title page.
latex_logo = 'images/nngt_logo.pdf'

# For "manual" documents, if this is true, then toplevel headings are parts,
# not chapters.
#latex_use_parts = False

# If true, show page references after internal links.
#latex_show_pagerefs = False

# If true, show URL addresses after external links.
#latex_show_urls = False

# Documents to append as an appendix to all manuals.
#latex_appendices = []

# If false, no module index is generated.
#latex_domain_indices = True


# -- Options for manual page output ---------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [
    (master_doc, 'nngt', u'NNGT Documentation',
     [author], 1)
]

# If true, show URL addresses after external links.
#man_show_urls = False


# -- Options for Texinfo output -------------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
  (master_doc, 'NNGT', u'NNGT Documentation',
   author, 'NNGT', 'One line description of project.',
   'Miscellaneous'),
]

# Documents to append as an appendix to all manuals.
#texinfo_appendices = []

# If false, no module index is generated.
#texinfo_domain_indices = True

# How to display URL addresses: 'footnote', 'no', or 'inline'.
texinfo_show_urls = 'inline'

# If true, do not generate a @detailmenu in the "Top" node's menu.
#texinfo_no_detailmenu = False

autodoc_default_flags = ['members', 'undoc-members']
autodoc_docstring_signature = True
autosummary_generate = True
autodoc_order = 'bysource'
autoclass_content = 'both'
napoleon_include_special_with_doc = False
napoleon_use_param = False
napoleon_use_rtype = False
imported_members = True

intersphinx_mapping = {
    'cartopy': ('https://scitools.org.uk/cartopy/docs/latest/', None),
    'geopandas': ('https://geopandas.org/', None),
    'gt': ('https://graph-tool.skewed.de/static/doc/', None),
    'ipython': ('https://ipython.org/ipython-doc/stable/', None),
    'matplotlib': ('https://matplotlib.org/', None),
    'networkx': ('https://networkx.github.io/documentation/stable/', None),
    'numpy': ('https://docs.scipy.org/doc/numpy', None),
    'python': ('https://docs.python.org/3/', None),
    'scipy': ('https://docs.scipy.org/doc/scipy/reference', None),
    'shapely': ('https://shapely.readthedocs.io/en/latest/', None),
}

extlinks_fancy = {
    'doi': (['https://dx.doi.org/{0}'], ['DOI: {0}']),
    'arxiv': (['https://arxiv.org/abs/{0}'], ['arXiv: {0}']),
    'gtdoc': (['https://graph-tool.skewed.de/static/doc/{0}.html#graph_tool.{1}'], ['graph-tool - {0}']),
    'igdoc': (['https://igraph.org/python/doc/igraph.GraphBase-class.html#{0}'], ['igraph - {0}']),
    'nxdoc': (['https://networkx.github.io/documentation/stable/reference/{0}generated/networkx.{1}.html'], ['networkx - {0}'])
}

# sphinx gallery parameters

sphinx_gallery_conf = {
     'examples_dirs': ['examples/graph_structure', 'examples/graph_properties'],  # path to your example scripts
     'gallery_dirs': ['gallery/graph_structure', 'gallery/graph_properties'],  # path to where to save gallery generated output
     'thumbnail_size': (400, 400),
     'capture_repr': (),
}



#!/usr/bin/env python
#-*- coding:utf-8 -*-

# test_weights.py

"""
Test weights setting.
"""

import numpy as np
import pytest

import nngt
import nngt.generation as ng


@pytest.mark.mpi_skip
def test_set_weights():
    w = 10.
    g = ng.erdos_renyi(nodes=100, density=0.1, weights=w)

    assert set(g.get_weights()) == {w}

    w2 = 5.
    g.set_weights(w2)
    
    assert set(g.get_weights()) == {w2}

    elist = g.get_edges()[:10]  # keep 10 first edges

    w3 = 2.
    g.set_weights(w3, elist=elist)

    assert set(g.get_weights()) == {w2, w3}
    assert set(g.get_weights(edges=elist)) == {w3}


def test_weighted_degrees():
    g = nngt.Graph()
    g.new_node(10)

    ww = [5.]*4
    g.new_edges([(0, 4), (8, 2), (7, 0), (1, 3)], attributes={"weight": ww})

    # check degrees
    in_deg  = [1., 0., 1., 1., 1., 0., 0., 0., 0., 0.]
    out_deg = [1., 1., 0., 0., 0., 0., 0., 1., 1., 0.]
    tot_deg = [2., 1., 1., 1., 1., 0., 0., 1., 1., 0.]

    assert np.all(g.get_degrees("in") == in_deg)
    assert np.all(
        g.get_degrees("in", weights=True) == np.multiply(ww[0], in_deg))
    assert np.all(g.get_degrees("out") == out_deg)
    assert np.all(g.get_degrees("total") == tot_deg)

    # set all weights to zero
    g.set_weights(0.)

    assert set(g.get_weights()) == {0.}
    assert set(g.get_edge_attributes(name="weight")) == {0.}

    assert not np.any(g.get_degrees("in", weights=True))

    # set two edges to 2.
    elist = [(0, 4), (8, 2)]
    g.set_weights(2., elist=elist)

    assert set(g.get_weights()) == {0., 2.}
    assert set(g.get_edge_attributes(name="weight")) == {0., 2.}

    # test new weighted degree
    w_indeg  = [0., 0., 2., 0., 2., 0., 0., 0., 0., 0.]
    w_outdeg = [2., 0., 0., 0., 0., 0., 0., 0., 2., 0.]

    assert np.all(g.get_degrees("in", weights=True) == w_indeg)

    assert np.all(g.get_degrees("out", weights=True) == w_outdeg)

    # check the node and edges
    assert g.node_nb() == 10
    assert g.edge_nb() == 4


if __name__ == "__main__":
    test_set_weights()
    test_weighted_degrees()



#-*- coding:utf-8 -*-
#
# analysis/nx_functions.py
#
# This file is part of the NNGT project, a graph-library for standardized and
# and reproducible graph analysis: generate and analyze networks with your
# favorite graph library (graph-tool/igraph/networkx) on any platform, without
# any change to your code.
# Copyright (C) 2015-2021 Tanguy Fardet
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

""" Tools to analyze graphs with the networkx backend """

import numpy as np
import scipy.sparse as ssp

from ..lib.test_functions import nonstring_container, is_integer
from ..lib.graph_helpers import _get_nx_weights, _get_nx_graph

import networkx as nx


def global_clustering_binary_undirected(g):
    '''
    Returns the undirected global clustering coefficient.

    This corresponds to the ratio of undirected triangles to the number of
    undirected triads.

    Parameters
    ----------
    g : :class:`~nngt.Graph`
        Graph to analyze.

    References
    ----------
    .. [nx-global-clustering] :nxdoc:`algorithms.cluster.transitivity`
    '''
    return nx.transitivity(g.graph.to_undirected(as_view=True))


def local_clustering_binary_undirected(g, nodes=None):
    '''
    Returns the undirected local clustering coefficient of some `nodes`.

    If `g` is directed, then it is converted to a simple undirected graph
    (no parallel edges).

    Parameters
    ----------
    g : :class:`~nngt.Graph`
        Graph to analyze.
    nodes : list, optional (default: all nodes)
        The list of nodes for which the clustering will be returned

    Returns
    -------
    lc : :class:`numpy.ndarray`
        The list of clustering coefficients, on per node.

    References
    ----------
    .. [nx-local-clustering] :nxdoc:`algorithms.cluster.clustering`
    '''
    num_nodes = g.node_nb()

    if nonstring_container(nodes):
        num_nodes = len(nodes)
    elif nodes is not None:
        num_nodes = 1

    lc = nx.clustering(g.graph.to_undirected(as_view=True), nodes=nodes,
                       weight=None)

    if num_nodes == 1:
        return lc

    if nodes is None:
        nodes = list(range(num_nodes))

    return np.array([lc[n] for n in nodes], dtype=float)


def assortativity(g, degree, weights=None):
    '''
    Returns the assortativity of the graph.
    This tells whether nodes are preferentially connected together depending
    on their degree.

    Parameters
    ----------
    g : :class:`~nngt.Graph`
        Graph to analyze.
    degree : str
        The type of degree that should be considered.
    weights : bool or str, optional (default: binary edges)
        Whether edge weights should be considered; if ``None`` or ``False``
        then use binary edges; if ``True``, uses the 'weight' edge attribute,
        otherwise uses any valid edge attribute required.

    References
    ----------
    .. [nx-assortativity]
       :nxdoc:`algorithms.assortativity.degree_assortativity_coefficient`
    '''
    if weights is not None:
        raise NotImplementedError("Weighted assortatibity is not yet "
                                  "implemented for networkx backend.")

    w = _get_nx_weights(g, weights)

    return nx.degree_assortativity_coefficient(g.graph, x=degree, y=degree,
                                               weight=w)


def reciprocity(g):
    '''
    Calculate the edge reciprocity of the graph.

    The reciprocity is defined as the number of edges that have a reciprocal
    edge (an edge between the same nodes but in the opposite direction)
    divided by the total number of edges.
    This is also the probability for any given edge, that its reciprocal edge
    exists.
    By definition, the reciprocity of undirected graphs is 1.

    @todo: check whether we can get this for single nodes for all libraries.

    Parameters
    ----------
    g : :class:`~nngt.Graph`
        Graph to analyze.

    References
    ----------
    .. [nx-reciprocity] :nxdoc:`algorithms.reciprocity.overall_reciprocity`
    '''
    if not g.is_directed():
        return 1.

    return nx.overall_reciprocity(g.graph)


def closeness(g, weights=None, nodes=None, mode="out", harmonic=False,
              default=np.NaN):
    r'''
    Returns the closeness centrality of some `nodes`.

    Closeness centrality of a node `u` is defined, for the harmonic version,
    as the sum of the reciprocal of the shortest path distance :math:`d_{uv}`
    from `u` to the N - 1 other nodes in the graph (if `mode` is "out",
    reciprocally :math:`d_{vu}`, the distance to `u` from another node v,
    if `mode` is "in"):

    .. math::

        C(u) = \frac{1}{N - 1} \sum_{v \neq u} \frac{1}{d_{uv}},

    or, using the arithmetic definition, as the reciprocal of the
    average shortest path distance to/from `u` over to all other nodes:

    .. math::

        C(u) = \frac{n - 1}{\sum_{v \neq u} d_{uv}},

    where `d_{uv}` is the shortest-path distance from `u` to `v`,
    and `n` is the number of nodes in the component.

    By definition, the distance is infinite when nodes are not connected by
    a path in the harmonic case (such that :math:`\frac{1}{d(v, u)} = 0`),
    while the distance itself is taken as zero for unconnected nodes in the
    first equation.

    Parameters
    ----------
    g : :class:`~nngt.Graph`
        Graph to analyze.
    weights : bool or str, optional (default: binary edges)
        Whether edge weights should be considered; if ``None`` or ``False``
        then use binary edges; if ``True``, uses the 'weight' edge attribute,
        otherwise uses any valid edge attribute required.
    nodes : list, optional (default: all nodes)
        The list of nodes for which the clutering will be returned
    mode : str, optional (default: "out")
        For directed graphs, whether the distances are computed from ("out") or
        to ("in") each of the nodes.
    harmonic : bool, optional (default: False)
        Whether the arithmetic (default) or the harmonic (recommended) version
        of the closeness should be used.

    Returns
    -------
    c : :class:`numpy.ndarray`
        The list of closeness centralities, on per node.

    .. warning ::
        For compatibility reasons (harmonic closeness is not implemented for
        igraph), the arithmetic version is used by default; however, it is
        recommended to use the harmonic version instead whenever possible.

    References
    ----------
    .. [nx-harmonic] :nxdoc:`algorithms.centrality.harmonic_centrality`
    .. [nx-closeness] :nxdoc:`algorithms.centrality.closeness_centrality`
    '''
    w = _get_nx_weights(g, weights)

    graph = g.graph

    if graph.is_directed() and mode == "out":
        graph = g.graph.reverse(copy=False)

    c = None

    if harmonic:
        c = nx.harmonic_centrality(graph, distance=w)
    else:
        c = nx.closeness_centrality(graph, distance=w, wf_improved=False)

    c = np.array([v for _, v in c.items()])

    # normalize
    if harmonic:
        c *= 1 / (len(graph) - 1)
    elif default != 0:
        c[c == 0.] = default

    if nodes is None:
        return c

    return c[nodes]


def betweenness(g, btype="both", weights=None):
    '''
    Returns the normalized betweenness centrality of the nodes and edges.

    Parameters
    ----------
    g : :class:`~nngt.Graph`
        Graph to analyze.
    btype : str, optional (default 'both')
        The centrality that should be returned (either 'node', 'edge', or
        'both'). By default, both betweenness centralities are computed.
    weights : bool or str, optional (default: binary edges)
        Whether edge weights should be considered; if ``None`` or ``False``
        then use binary edges; if ``True``, uses the 'weight' edge attribute,
        otherwise uses any valid edge attribute required.

    Returns
    -------
    nb : :class:`numpy.ndarray`
        The nodes' betweenness if `btype` is 'node' or 'both'
    eb : :class:`numpy.ndarray`
        The edges' betweenness if `btype` is 'edge' or 'both'

    References
    ----------
    .. [nx-ebetw] :nxdoc:`algorithms.centrality.edge_betweenness_centrality`
    .. [nx-nbetw] :nxdoc:`networkx.algorithms.centrality.betweenness_centrality`
    '''
    w = _get_nx_weights(g, weights)

    nb, eb = None, None

    if btype in ("both", "node"):
        di_nb = nx.betweenness_centrality(g.graph, weight=w)
        nb    = np.array([di_nb[i] for i in g.get_nodes()])

    if btype in ("both", "edge"):
        di_eb = nx.edge_betweenness_centrality(g.graph, weight=w)
        eb    = np.array([di_eb[tuple(e)] for e in g.edges_array])

    if btype == "node":
        return nb
    elif btype == "edge":
        return eb

    return nb, eb


def connected_components(g, ctype=None):
    '''
    Returns the connected component to which each node belongs.

    Parameters
    ----------
    g : :class:`~nngt.Graph`
        Graph to analyze.
    ctype : str, optional (default 'scc')
        Type of component that will be searched: either strongly connected
        ('scc', by default) or weakly connected ('wcc').

    Returns
    -------
    cc, hist : :class:`numpy.ndarray`
        The component associated to each node (`cc`) and the number of nodes in
        each of the component (`hist`).

    References
    ----------
    .. [nx-ucc] :nxdoc:`algorithms.components.connected_components`
    .. [nx-scc] :nxdoc:`algorithms.components.strongly_connected_components`
    .. [nx-wcc] :nxdoc:`algorithms.components.weakly_connected_components`
    '''
    if ctype is None:
        ctype = "scc" if g.is_directed() else "wcc"

    res   = None

    if not g.is_directed():
        res = nx.connected_components(g.graph)
    elif ctype == "scc":
        res = nx.strongly_connected_components(g.graph)
    elif ctype == "wcc":
        res = nx.weakly_connected_components(g.graph)
    else:
        raise ValueError("Invalid `ctype`, only 'scc' and 'wcc' are allowed.")

    cc   = np.zeros(g.node_nb(), dtype=int)
    hist = []

    for i, nodes in enumerate(res):
        cc[list(nodes)] = i

        hist.append(len(nodes))

    return cc, np.array(hist, dtype=int)


def shortest_path(g, source, target, directed=None, weights=None,
                  combine_weights="mean"):
    '''
    Returns a shortest path between `source`and `target`.
    The algorithms returns an empty list if there is no path between the nodes.

    Parameters
    ----------
    g : :class:`~nngt.Graph`
        Graph to analyze.
    source : int
        Node from which the path starts.
    target : int
        Node where the path ends.
    directed : bool, optional (default: ``g.is_directed()``)
        Whether the edges should be considered as directed or not
        (automatically set to False if `g` is undirected).
    weights : str or array, optional (default: binary)
        Whether to use weighted edges to compute the distances. By default,
        all edges are considered to have distance 1.
    combine_weights : str, optional (default: 'mean')
        How to combine the weights of reciprocal edges if the graph is directed
        but `directed` is set to False. It can be:

        * "sum": the sum of the edge attribute values will be used for the new
          edge.
        * "mean": the mean of the edge attribute values will be used for the
          new edge.
        * "min": the minimum of the edge attribute values will be used for the
          new edge.
        * "max": the maximum of the edge attribute values will be used for the
          new edge.

    Returns
    -------
    path : list of ints
        Order of the nodes making up the path from `source` to `target`.

    References
    ----------
    .. [nx-sp] :nxdoc:`algorithms.shortest_paths.generic.shortest_path`
    '''
    g, graph, w = _get_nx_graph(g, directed, weights, combine_weights,
                             return_all=True)

    w = _get_nx_weights(g, w)

    try:
        return nx.shortest_path(graph, source, target, weight=w)
    except nx.NetworkXNoPath:
        return []


def all_shortest_paths(g, source, target, directed=None, weights=None,
                       combine_weights="mean"):
    '''
    Yields all shortest paths from `source` to `target`.
    The algorithms returns an empty generator if there is no path between the
    nodes.

    Parameters
    ----------
    g : :class:`~nngt.Graph`
        Graph to analyze.
    source : int
        Node from which the paths starts.
    target : int, optional (default: all nodes)
        Node where the paths ends.
    directed : bool, optional (default: ``g.is_directed()``)
        Whether the edges should be considered as directed or not
        (automatically set to False if `g` is undirected).
    weights : str or array, optional (default: binary)
        Whether to use weighted edges to compute the distances. By default,
        all edges are considered to have distance 1.
    combine_weights : str, optional (default: 'mean')
        How to combine the weights of reciprocal edges if the graph is directed
        but `directed` is set to False. It can be:

        * "sum": the sum of the edge attribute values will be used for the new
          edge.
        * "mean": the mean of the edge attribute values will be used for the
          new edge.
        * "min": the minimum of the edge attribute values will be used for the
          new edge.
        * "max": the maximum of the edge attribute values will be used for the
          new edge.

    Returns
    -------
    all_paths : generator
        Generator yielding paths as lists of ints.

    References
    ----------
    .. [nx-sp] :nxdoc:`algorithms.shortest_paths.generic.all_shortest_paths`
    '''
    g, graph, w = _get_nx_graph(g, directed, weights, combine_weights,
                             return_all=True)

    w = _get_nx_weights(g, w)

    try:
        return nx.all_shortest_paths(graph, source, target, weight=w)
    except nx.NetworkXNoPath:
        return (_ for _ in [])


def shortest_distance(g, sources=None, targets=None, directed=None,
                      weights=None, combine_weights="mean"):
    '''
    Returns the length of the shortest paths between `sources`and `targets`.
    The algorithms return infinity if there are no paths between nodes.

    Parameters
    ----------
    g : :class:`~nngt.Graph`
        Graph to analyze.
    sources : list of nodes, optional (default: all)
        Nodes from which the paths must be computed.
    targets : list of nodes, optional (default: all)
        Nodes to which the paths must be computed.
    directed : bool, optional (default: ``g.is_directed()``)
        Whether the edges should be considered as directed or not
        (automatically set to False if `g` is undirected).
    weights : str or array, optional (default: binary)
        Whether to use weighted edges to compute the distances. By default,
        all edges are considered to have distance 1.
    combine_weights : str, optional (default: 'mean')
        How to combine the weights of reciprocal edges if the graph is directed
        but `directed` is set to False. It can be:

        * "sum": the sum of the edge attribute values will be used for the new
          edge.
        * "mean": the mean of the edge attribute values will be used for the
          new edge.
        * "min": the minimum of the edge attribute values will be used for the
          new edge.
        * "max": the maximum of the edge attribute values will be used for the
          new edge.

    Returns
    -------
    distance : float, or 1d/2d numpy array of floats
        Distance (if single source and single target) or distance array.
        For multiple sources and targets, the shape of the matrix is (S, T),
        with S the number of sources and T the number of targets; for a single
        source or target, return a 1d-array of length T or S.

    References
    ----------
    .. [nx-sp] :nxdoc:`algorithms.shortest_paths.weighted.multi_source_dijkstra`
    '''
    num_nodes = g.node_nb()

    # check consistency for weights and directed
    g, graph, w = _get_nx_graph(g, directed, weights, combine_weights,
                             return_all=True)

    w = _get_nx_weights(g, w)

    # check for single source/target case and convert sources and targets
    if is_integer(sources):
        if is_integer(targets):
            try:
                return nx.shortest_path_length(graph, sources, targets,
                                               weight=w)
            except Exception as e:
                return np.inf

        sources = [sources]
    elif sources is None:
        sources = range(num_nodes)

    if is_integer(targets):
        targets = [targets]

    # compute distances
    data, ii, jj = [], [], []

    def _nx_sp(nx_graph, s, weight):
        if weight is None:
            return nx.single_source_shortest_path_length(nx_graph, s)

        dist, _ = nx.multi_source_dijkstra(graph, [s], weight=weight)

        return dist

    for s in sources:
        dist = _nx_sp(graph, s, w)

        if targets is None:
            data.extend(dist.values())
            ii.extend((s for _ in range(len(dist))))
            jj.extend(dist.keys())
        else:
            for t in targets:
                if t in dist:
                    data.append(dist[t])
                    ii.append(s)
                    jj.append(t)

    num_sources = num_nodes if sources is None else len(sources)
    num_targets = num_nodes if targets is None else len(targets)

    mat_dist = np.full((num_sources, num_targets), np.inf)
    mat_dist[ii, jj] = data

    if num_sources == 1:
        return mat_dist[0]

    if num_targets == 1:
        return mat_dist.T[0]

    return mat_dist


def average_path_length(g, sources=None, targets=None, directed=None,
                        weights=None, combine_weights="mean",
                        unconnected=False):
    r'''
    Returns the average shortest path length between `sources` and `targets`.
    The algorithms raises an error if all nodes are not connected unless
    `unconnected` is set to True.

    The average path length is defined as

    .. math::

       L = \frac{1}{N_p} \sum_{u,v} d(u, v),

    where :math:`N_p` is the number of paths between `sources` and `targets`,
    and :math:`d(u, v)` is the shortest path distance from u to v.

    If `sources` and `targets` are both None, then the total number of paths is
    :math:`N_p = N(N - 1)`, with :math:`N` the number of nodes in the graph.

    Parameters
    ----------
    g : :class:`~nngt.Graph`
        Graph to analyze.
    sources : list of nodes, optional (default: all)
        Nodes from which the paths must be computed.
    targets : list of nodes, optional (default: all)
        Nodes to which the paths must be computed.
    directed : bool, optional (default: ``g.is_directed()``)
        Whether the edges should be considered as directed or not
        (automatically set to False if `g` is undirected).
    weights : str or array, optional (default: binary)
        Whether to use weighted edges to compute the distances. By default,
        all edges are considered to have distance 1.
    combine_weights : str, optional (default: 'mean')
        How to combine the weights of reciprocal edges if the graph is directed
        but `directed` is set to False. It can be:

        * "sum": the sum of the edge attribute values will be used for the new
          edge.
        * "mean": the mean of the edge attribute values will be used for the
          new edge.
        * "min": the minimum of the edge attribute values will be used for the
          new edge.
        * "max": the maximum of the edge attribute values will be used for the
          new edge.
    unconnected : bool, optional (default: False)
        If set to true, ignores unconnected nodes and returns the average path
        length of the existing paths.

    References
    ----------
    .. [nx-sp] :nxdoc:`algorithms.shortest_paths.generic.average_shortest_path_length`
    '''
    directed = g.is_directed() if directed is None else directed

    if sources is None and targets is None and not unconnected:
        g, graph, w = _get_nx_graph(g, directed, weights, combine_weights,
                                 return_all=True)

        w = _get_nx_weights(g, w)

        return nx.average_shortest_path_length(graph, weight=w)

    mat_dist = shortest_distance(g, sources=sources, targets=targets,
                                 directed=directed, weights=weights)

    if not unconnected and np.any(np.isinf(mat_dist)):
        raise nx.NetworkXNoPath("`sources` and `target` do not belong to the "
                                "same connected component.")

    # compute the number of path
    num_paths = np.sum(mat_dist != 0)

    # compute average path length
    if unconnected:
        num_paths -= np.sum(np.isinf(mat_dist))

        return np.nansum(mat_dist) / num_paths

    return np.sum(mat_dist) / num_paths


def diameter(g, directed=None, weights=None, combine_weights="mean",
             is_connected=False):
    '''
    Returns the diameter of the graph.

    .. versionchanged:: 2.3
        Added `combine_weights` argument.

    .. versionchanged:: 2.0
        Added `directed` and `is_connected` arguments.

    It returns infinity if the graph is not connected (strongly connected for
    directed graphs) unless `is_connected` is True, in which case it returns
    the longest existing shortest distance.

    Parameters
    ----------
    g : :class:`~nngt.Graph`
        Graph to analyze.
    directed : bool, optional (default: ``g.is_directed()``)
        Whether to compute the directed diameter if the graph is directed.
        If False, then the graph is treated as undirected. The option switches
        to False automatically if `g` is undirected.
    weights : bool or str, optional (default: binary edges)
        Whether edge weights should be considered; if ``None`` or ``False``
        then use binary edges; if ``True``, uses the 'weight' edge attribute,
        otherwise uses any valid edge attribute required.
    combine_weights : str, optional (default: 'mean')
        How to combine the weights of reciprocal edges if the graph is directed
        but `directed` is set to False. It can be:

        * "sum": the sum of the edge attribute values will be used for the new
          edge.
        * "mean": the mean of the edge attribute values will be used for the
          new edge.
        * "min": the minimum of the edge attribute values will be used for the
          new edge.
        * "max": the maximum of the edge attribute values will be used for the
          new edge.
    is_connected : bool, optional (default: False)
        If False, check whether the graph is connected or not and return
        infinite diameter if graph is unconnected. If True, the graph is
        assumed to be connected.

    See also
    --------
    :func:`nngt.analysis.shortest_distance`

    References
    ----------
    .. [nx-diameter] :nxdoc:`algorithms.distance_measures.diameter`
    .. [nx-dijkstra] :nxdoc:`algorithms.shortest_paths.weighted.all_pairs_dijkstra`
    '''
    w = _get_nx_weights(g, weights)

    # weighted or "connected" cases
    if w is not None or is_connected:
        dist = shortest_distance(g, directed=directed, weights=weights,
                                 combine_weights=combine_weights)

        if is_connected:
            return np.max(dist[~np.isinf(dist)])

        return np.max(dist)

    # unweighted case
    graph = _get_nx_graph(g, directed, w, combine_weights)

    try:
        return nx.diameter(graph)
    except nx.exception.NetworkXError:
        return np.inf


def adj_mat(g, weights=None, mformat="csr"):
    r'''
    Returns the adjacency matrix :math:`A` of the graph.
    With edge :math:`i \leftarrow j` corresponding to entry :math:`A_{ij}`.

    Parameters
    ----------
    g : :class:`~nngt.Graph`
        Graph to analyze.
    weights : bool or str, optional (default: binary edges)
        Whether edge weights should be considered; if ``None`` or ``False``
        then returns the binary adjacency matrix; if ``True``, returns the
        weighted matrix, otherwise fills the matrix with any valid edge
        attribute values.
    mformat : str, optional (default: "csr")
        Type of :mod:`scipy.sparse` matrix that will be returned, by
        default :class:`scipy.sparse.csr_matrix`.

    Returns
    -------
    The adjacency matrix as a :class:`scipy.sparse.csr_matrix`.

    References
    ----------
    .. [nx-adjacency] :nxdoc:`.convert_matrix.to_scipy_sparse_matrix`
    '''
    w = _get_nx_weights(g, weights)

    return nx.to_scipy_sparse_matrix(g.graph, weight=w, format=mformat)


def get_edges(g):
    '''
    Returns the edges in the graph by order of creation.
    '''
    return g.edges_array



"""
Example file
"""

import os
import sys
sys.path.append(os.path.abspath(".."))

import matplotlib.pyplot as plt
import numpy as np

from mpl_chord_diagram import chord_diagram

# flux matrix

flux = np.array([
    [11975,  5871, 8916, 2868],
    [ 1951, 10048, 2060, 6171],
    [ 8010, 16145, 81090, 8045],
    [ 1013,   990,  940, 6907]
])

names = ['non-crystal', 'FCC', 'HCP', 'BCC']

# plot different examples

grads = (True, False, False, True)                # gradient
gaps  = (0.03, 0, 0.03, 0)                        # gap value
sorts = ("size", "size", "distance", "distance")  # sort type
cclrs = (None, None, "slategrey", None)           # chord colors
nrota = (False, False, True, True)                # name rotation
cmaps = (None, None, None, "summer")              # colormap
fclrs = "grey"                                    # fontcolors

for grd, gap, srt, cc, nr, cm in zip(grads, gaps, sorts, cclrs, nrota, cmaps):
    chord_diagram(flux, names, gap=gap, use_gradient=grd, sort=srt,
                  cmap=cm, chord_colors=cc, rotate_names=nr, fontcolor=fclrs)

    str_grd = "_gradient" if grd else ""

    plt.savefig(
        "images/example{}_sort-{}.png".format(str_grd, srt),
                dpi=600, transparent=True, bbox_inches='tight',
                pad_inches=0.02)

plt.show()



#-*- coding:utf-8 -*-
#
# geospatial/countries.py
#
# This file is part of the NNGT project, a graph-library for standardized and
# and reproducible graph analysis: generate and analyze networks with your
# favorite graph library (graph-tool/igraph/networkx) on any platform, without
# any change to your code.
# Copyright (C) 2015-2021 Tanguy Fardet
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

import os

import numpy as np

import cartopy
import geopandas as gpd

from ._cartopy_ne import natural_earth
from shapely.geometry import Point


# --------------------------------------------- #
# Load and store the worlds at different scales #
# --------------------------------------------- #

try:
    fname_110 = natural_earth(resolution='110m', category='cultural',
                              name='admin_0_countries')

    fname_50 = natural_earth(resolution='50m', category='cultural',
                             name='admin_0_countries')

    fname_10 = natural_earth(resolution='10m', category='cultural',
                             name='admin_0_map_units')

    fname_cities = natural_earth(resolution='10m', category='cultural',
                                 name='populated_places_simple')
except Exception as e:
    print("Error excountered:", e)
    raise RuntimeError("The `geospatial` module needs to download the "
                       "NaturalEarth maps on its first run, please make sure "
                       "you are connected to the internet or manually add the "
                       "files to '{}'".format(cartopy.config['data_dir']))

world_110 = gpd.read_file(fname_110)
world_50  = gpd.read_file(fname_50)
world_10  = gpd.read_file(fname_10)
cities    = gpd.read_file(fname_cities)

maps = {
    "110m": world_110,
    "50m": world_50,
    "10m": world_10,
    "cities": cities,
}


# ----------------------------------------- #
# Store countries and their world/positions #
# ----------------------------------------- #

country_codes_adaptive = {}
country_codes_110 = {}
country_codes_50 = {}
country_codes_10 = {}

country_names_adaptive = {}
country_names_110 = {}
country_names_50 = {}
country_names_10 = {}

ctn_adaptive = {}
ctn_110 = {}
ctn_50 = {}
ctn_10 = {}

countries = {"Country", "Sovereign country", 'Geo unit'}

codes = ("GU_A3", "SOV_A3", "ADM0_A3")

for i, v in world_110.iterrows():
    name = v.NAME_LONG
    code = v.SU_A3

    cvalues = [v[c] for c in codes]

    country_names_110[name] = i
    country_names_adaptive[name] = i

    country_codes_110[code] = i
    country_codes_adaptive[code] = i

    ctn_110[code] = name
    ctn_adaptive[code] = name

    for cval in cvalues:
        if v.TYPE in countries and cval not in country_codes_adaptive:
            country_codes_110[cval] = i
            country_codes_adaptive[cval] = i
            ctn_110[cval] = name
            ctn_adaptive[cval] = name


new_countries = set()
size_adaptive = len(world_110)

for i, v in world_50.iterrows():
    name = v.NAME_LONG
    code = v.SU_A3

    cvalues = [v[c] for c in codes]

    country_names_50[name] = i
    country_codes_50[code] = i

    ctn_50[code] = name

    if name not in country_names_adaptive:
        country_names_adaptive[name] = len(new_countries) + size_adaptive
        country_codes_adaptive[code] = len(new_countries) + size_adaptive
        ctn_adaptive[code] = name

        new_countries.add(i)

    for cval in cvalues:
        if v.TYPE in countries and cval not in country_codes_50:
            country_codes_50[cval] = i
            ctn_50[cval] = name

            if cval not in country_codes_adaptive:
                country_codes_adaptive[cval] = \
                    len(new_countries) + size_adaptive
                ctn_adaptive[cval] = name

world = world_110.append(world_50.iloc[list(new_countries)], ignore_index=True)

new_countries = []
size_adaptive = len(world)

for i, v in world_10.iterrows():
    name = v.NAME_LONG
    code = v.SU_A3

    cvalues = [v[c] for c in codes]

    country_names_10[name] = i
    country_codes_10[code] = i

    ctn_10[code] = name

    for cval in cvalues:
        if v.TYPE in countries and cval not in country_codes_10:
            country_codes_10[cval] = i
            ctn_10[cval] = name

    if name not in country_names_adaptive:
        sovc = v.SOV_A3

        # check if it's a subunit, if so, add only if it does not overlap
        # with sovereign territory (special case for Antigua and Barbuda)
        if v.LEVEL == 3 and sovc != "ATG":
            geom = v.geometry

            # get soverign territory
            idx = country_codes_adaptive[sovc]

            sov_geom = world.iloc[idx].geometry

            overlap = geom.intersection(sov_geom).area > 0.1*geom.area

            if overlap:
                continue

        country_names_adaptive[name] = len(new_countries) + size_adaptive
        country_codes_adaptive[code] = len(new_countries) + size_adaptive

        ctn_adaptive[code] = name

        if v.TYPE in countries:
            for cval in cvalues:
                if cval not in country_codes_adaptive:
                    country_codes_adaptive[cval] = \
                        len(new_countries) + size_adaptive
                    ctn_adaptive[cval] = name

        new_countries.append(i)


world = world.append(world_10.iloc[new_countries], ignore_index=True)

maps["adaptive"] = world

country_names = {
    "110m": country_names_110,
    "50m": country_names_50,
    "10m": country_names_10,
    "adaptive": country_names_adaptive
}

country_codes = {
    "110m": country_codes_110,
    "50m": country_codes_50,
    "10m": country_codes_10,
    "adaptive": country_codes_adaptive
}

codes_to_names = {
    "110m": ctn_110,
    "50m": ctn_50,
    "10m": ctn_10,
    "adaptive": ctn_adaptive
}


# ------------------------------------- #
# Add usual covertors for country names #
# ------------------------------------- #

convertors = {
    'Democratic Republic of Korea': 'Dem. Rep. Korea',
    'Iran (Islamic Republic of)': 'Iran',
    'Venezuela (Bolivarian Republic of)': 'Venezuela',
    'Bolivia (Plurinational State of)': 'Bolivia',
    'China, Taiwan Province of': 'Taiwan',
    'North Macedonia': 'Macedonia',
    'China, mainland': 'China',
    'Serbia and Montenegro': 'Serbia',
    'Sudan (former)': 'Sudan',
    'Belgium-Luxembourg': 'Belgium',
    'Ethiopia PDR': 'Ethiopia',
    'Yugoslav SFR': 'Serbia',
    'Cabo Verde': 'Republic of Cabo Verde',
    'Eswatini': 'eSwatini',
    'Sao Tome and Principe': 'São Tomé and Principe',
    'China, Hong Kong SAR': 'Hong Kong',
    'China, Macao SAR': 'Macao',
    'Congo': 'Republic of the Congo',
    'Czechia': 'Czech Republic',
    'Gambia': 'The Gambia',
    'Netherlands Antilles (former)': 'Curaçao',
    'Falkland Islands (Malvinas)': 'Falkland Islands',
    'Viet Nam': 'Vietnam',
    'Saint Helena, Ascension and Tristan da Cunha': 'Saint Helena',
    'Pitcairn': 'Pitcairn Islands',
    'Micronesia (Federated States of)': 'Federated States of Micronesia',
    'French Southern Territories': 'French Southern and Antarctic Lands',
    'Palestinian Territory': 'Palestine'
}


for k, v in zip(world.FORMAL_EN, world.NAME_LONG):
    if k is not None:
        convertors[k] = v


convertors.update(
    {k: v for k, v in zip(world.NAME_EN, world.NAME_LONG)})


# ----------------------------------- #
# Representative points for countries #
# ----------------------------------- #

points = []

for _, v in world.iterrows():
    if v.TYPE in countries:
        # check capital
        idx = np.where((cities.sov_a3 == v.SOV_A3)*cities.adm0cap)[0]

        if len(idx) == 1:
            points.append(cities.iloc[idx[0]].geometry)
            continue
        elif len(idx) > 1:
            pop = cities.iloc[idx].pop_max
            idx = idx[np.argmax(pop)]
            points.append(cities.iloc[idx].geometry)
            continue
        else:
            idx = np.where((cities.sov_a3 == v.SU_A3)*cities.adm0cap)[0]

            if len(idx) == 1:
                points.append(cities.iloc[idx[0]].geometry)
                continue
            elif len(idx) > 1:
                pop = cities.iloc[idx].pop_max
                idx = idx[np.argmax(pop)]
                points.append(cities.iloc[idx].geometry)
                continue

    points.append(v.geometry.representative_point())
        

country_points = gpd.GeoDataFrame({
    'country': world.NAME_LONG,
    'geocode': world.SU_A3,
    'geometry': points
})

country_points.set_crs(epsg=world.crs.to_epsg(), inplace=True)



#!/usr/bin/env python
#-*- coding:utf-8 -*-
#
# This file is part of the NNGT project to generate and analyze
# neuronal networks and their activity.
# Copyright (C) 2015-2019  Tanguy Fardet
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

''' Introduction to neural groups '''

import numpy as np

import nngt
from nngt import Group, MetaGroup, Structure
from nngt import MetaNeuralGroup, NeuralGroup, NeuralPop


''' ------------- #
# Creating groups #
# ------------- '''

# the default group is empty, which is not very useful in general
empty_group = nngt.Group()
print(empty_group)
print("Group is empty?", empty_group.size == 0, "\nIt has therefore no ids:",
      empty_group.ids, "\n")

# to create a useful group, one can just say how many nodes it should contain
group1 = Group(500)  # a group with 500 nodes
print("Ids are not created", group1.ids, "len(group1) is", len(group1),
      "but the size is stored in group1.size:", group1.size, "\n")

# if you want to set the ids directly, you can pass them directly, otherwise
# they will be determine automatically when a Network is created with the group
group2 = NeuralGroup(range(10, 20))  # neurons with ids from 10 to 19
print("Neuron ids are:", group2.ids, "\n")


''' ------------------------- #
# Creating a structured graph #
# ------------------------- '''

room1 = nngt.Group(25)
room2 = nngt.Group(50)
room3 = nngt.Group(40)
room4 = nngt.Group(35)

names = ["R1", "R2", "R3", "R4"]

struct = nngt.Structure.from_groups((room1, room2, room3, room4), names)

g = nngt.Graph(structure=struct)

for room in struct:
    nngt.generation.connect_groups(g, room, room, "all_to_all")

nngt.generation.connect_groups(g, (room1, room2), struct, "erdos_renyi",
                               avg_deg=10, ignore_invalid=True)

nngt.generation.connect_groups(g, room3, room1, "erdos_renyi", avg_deg=20)

nngt.generation.connect_groups(g, room4, room3, "erdos_renyi", avg_deg=10)

if nngt.get_config("with_plot"):
    # chord diagram
    sg = g.get_structure_graph()

    nngt.plot.chord_diagram(sg, names="name", sort="distance",
                            use_gradient=True, show=True)

    # spring-block layout    
    nngt.plot.library_draw(g, node_cmap="viridis", show=True)


''' ------------------- #
# More group properties #
# ------------------- '''

# group can have names
named_group = Group(500, name="named_group")
print("I'm a named group!", named_group, "\n")

# NeuralGroups can store whether neurons are excitatory or inhibitory
exc   = NeuralGroup(800, neuron_type=1)   # excitatory group
exc2  = NeuralGroup(800, neuron_type=1)   # also excitatory
inhib = NeuralGroup(200, neuron_type=-1)  # inhibitory group
print("'exc2' is an excitatory group:", exc2.neuron_type == 1,
      "/ 'inhib' is an inhibitory group:", inhib.neuron_type == -1, "\n")


''' ---------------------------------- #
# Complete groups for NEST simulations #
# ---------------------------------- '''

# to make a complete neuronal group, one must include a valid neuronal type,
# model and (optionally) associated parameters

pyr = NeuralGroup(800, neuron_type=1, neuron_model="iaf_psc_alpha",
                  neuron_param={"tau_m": 50.}, name="pyramidal_cells")

fsi = NeuralGroup(200, neuron_type=-1, neuron_model="iaf_psc_alpha",
                  neuron_param={"tau_m": 20.},
                  name="fast_spiking_interneurons")


''' --------------------------- #
# Creating neuronal populations #
# --------------------------- '''

pop = NeuralPop.from_groups((pyr, fsi))

# making populations from scratch
pop = nngt.NeuralPop(with_models=False)              # empty population
pop.create_group(200, "first_group")                 # create excitatory group
pop.create_group(5, "second_group", neuron_type=-1)  # create inhibitory group

print("E/I population has size", pop.size, "and contains",
      len(pop), "groups:", pop.keys(), "\n")

# the two default populations
unif_pop = NeuralPop.uniform(1000)                     # only excitatory
ei_pop   = NeuralPop.exc_and_inhib(1000, iratio=0.25)  # 25% inhibitory

# check the groups inside
print("Uniform population has size", unif_pop.size, "and contains",
      len(unif_pop), "group:", unif_pop.keys(), "\n")
print("E/I population has size", ei_pop.size, "and contains",
      len(ei_pop), "groups:", ei_pop.keys(), "\n")

# A population can also be created from existing groups.
# Here we pass ``with_models=False`` to the population because these groups do
# not contain the information necessary to create a network in NEST (a valid
# neuron model).
print(exc.neuron_type, exc2.neuron_type, inhib.neuron_type)
ei_pop2 = NeuralPop.from_groups([exc, exc2, inhib], ["e1", "e2", "i"],
                                with_models=False)

print("E/I population has size", ei_pop2.size,
      "({} + {} + {}) and contains".format(exc.size, exc2.size, inhib.size),
      len(ei_pop2), "groups:", ei_pop2.keys(), "\n")


''' --------------------- #
# NEST-enabled population #
# --------------------- '''

# Let's create a population which will be used to make a network that can then
# be simulated with NEST.
# We create it from the pyramidal and fast spiking interneurons groups and
# add synaptic properties to the connections that will be made.
# (because the group already have names, we don't need to specify them again)

# optional synaptic properties
syn_spec = {
    'default': {"synaptic_model": "tsodyks2_synapse"},  # default connections
    ("pyramidal_cells", "pyramidal_cells"): {"U": 0.6}  # change a parameter
}

nest_pop = NeuralPop.from_groups([pyr, fsi], syn_spec=syn_spec)


''' ------------------------------- #
# Complex population and metagroups #
# ------------------------------- '''

# Let's model part of a cortical column with
# - granule cells in layer 2 and 4
# - pyramidal cells and interneurons in layers 3 and 5
# - indiscriminate cells in layer 6

nmod = "iaf_psc_exp"

idsL2gc = range(100)
idsL3py, idsL3i = range(100, 200), range(200, 300)
idsL4gc = range(300, 400)
idsL5py, idsL5i = range(400, 500), range(500, 600)
idsL6 = range(600, 700)

L2GC = NeuralGroup(idsL2gc, neuron_model=nmod, name="L2GC", neuron_type=1)
L3Py = NeuralGroup(idsL3py, neuron_model=nmod, name="L3Py", neuron_type=1)
L3I  = NeuralGroup(idsL3i,  neuron_model=nmod, name="L3I",  neuron_type=-1)
L4GC = NeuralGroup(idsL4gc, neuron_model=nmod, name="L4GC", neuron_type=1)
L5Py = NeuralGroup(idsL5py, neuron_model=nmod, name="L5Py", neuron_type=1)
L5I  = NeuralGroup(idsL5i,  neuron_model=nmod, name="L5I",  neuron_type=-1)
L6c  = NeuralGroup(idsL6,   neuron_model=nmod, name="L6c",  neuron_type=1)

# We can also group them by layers using metagroups (for L2/L3/L6 it is not
# really useful but it gives a coherent notation)
L2 = MetaGroup(idsL2gc, name="L2")
L3 = MetaNeuralGroup(L3Py.ids + L3I.ids, name="L3")
L4 = MetaGroup(idsL4gc, name="L4")
L5 = MetaNeuralGroup(L5Py.ids + L5I.ids, name="L5")
L6 = MetaGroup(idsL6, name="L6")

# Then we create the population from the groups
pop_column = NeuralPop.from_groups(
    [L2GC, L3Py, L3I, L4GC, L5Py, L5I, L6c], meta_groups=[L2, L3, L4, L5, L6])

# We can also add additional meta-groups for pyramidal, granule, and
# interneurons
pyr = MetaGroup(L3Py.ids + L5Py.ids, name="pyramidal")
pop_column.add_meta_group(pyr)  # add from existing meta-group

pop_column.create_meta_group(L3I.ids + L5I.ids, "interneurons")  # single line

pop_column.create_meta_group(L2GC.ids + L4GC.ids, "granule")

print("Column has meta-groups:", pop_column.meta_groups.keys(), "\n")

# Note that we used MetaNeuralGroups for L3 and L5 because they can separate
# excitatory and inhibitory neurons
print("Excitatory L3 neurons:", L3.excitatory, "\n")
print("Inhibitory L3 neurons:", L3.inhibitory)



#-*- coding:utf-8 -*-
#
# io/saving_helpers.py
#
# This file is part of the NNGT project, a graph-library for standardized and
# and reproducible graph analysis: generate and analyze networks with your
# favorite graph library (graph-tool/igraph/networkx) on any platform, without
# any change to your code.
# Copyright (C) 2015-2021 Tanguy Fardet
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

""" IO tools for NNGT """


def _neighbour_list(graph, separator, secondary, attributes):
    '''
    Generate a string containing the neighbour list of the graph as well as a
    dict containing the notifiers as key and the associated values.
    @todo: speed this up!
    '''
    lst_neighbours = list(graph.adjacency_matrix(mformat="lil").rows)

    for v1 in range(graph.node_nb()):
        for i, v2 in enumerate(lst_neighbours[v1]):
            str_edge = str(v2)
            eattr    = graph.get_edge_attributes((v1, v2))

            for attr in attributes:
                str_edge += "{}{}".format(secondary, eattr[attr])

            lst_neighbours[v1][i] = str_edge

        lst_neighbours[v1] = "{}{}{}".format(
            v1, separator, separator.join(lst_neighbours[v1]))

    str_neighbours = "\n".join(lst_neighbours)

    return str_neighbours


def _edge_list(graph, separator, secondary, attributes):
    ''' Generate a string containing the edge list and their properties. '''
    edges = graph.edges_array

    di_attributes = {
        k: v for k, v in graph.edge_attributes.items()
        if k != 'bweight'
    }

    end_strings = (len(attributes) - 1)*[secondary]

    end_strings.append('')

    lst_edges = []

    for i, e in enumerate(edges):
        str_edge = "{}{}{}".format(e[0], separator, e[1])

        if attributes:
            str_edge += separator

        for end, attr in zip(end_strings, attributes):
            str_edge += "{}{}".format(di_attributes[attr][i], end)

        lst_edges.append(str_edge)

    str_edges = "\n".join(lst_edges)

    return str_edges


def _dot(graph, attributes, **kwargs):
    pass


def _gml(graph, *args, **kwargs):
    ''' Generate a string containing the edge list and their properties. '''
    node_str = "  node\n  [\n    id {id}{attr}\n  ]"
    edge_str = "  edge\n  [\n    source {s}\n    target {t}\n{attr}\n  ]"
    attr_str = "    {key} {val}"

    indent = "    "

    # set nodes
    lst_elements = []

    for i in range(graph.node_nb()):
        lst_attr = []

        for k, v in graph.node_attributes.items():
            lst_attr.append(attr_str.format(key=k, val=v[i]))

        nattr = "\n" + "\n".join(lst_attr)

        lst_elements.append(node_str.format(id=i, attr=nattr))

    # set edges
    edges = graph.edges_array

    for i, e in enumerate(edges):
        lst_attr = []

        for k, v in graph.edge_attributes.items():
            lst_attr.append(attr_str.format(key=k, val=v[i]))

        eattr = "\n".join(lst_attr)

        lst_elements.append(edge_str.format(s=e[0], t=e[1], attr=eattr))

    str_gml = "\n".join(lst_elements)

    str_gml += "\n]"

    return str_gml


def _xml(graph, attributes, **kwargs):
    pass


def _gt(graph, attributes, **kwargs):
    pass


def _custom_info(graph_info, notifier, *args, **kwargs):
    ''' Format the graph information for custom formats '''
    info_str = ""

    for key, val in iter(graph_info.items()):
        info_str += "{}{}={}\n".format(notifier, key, val)

    return info_str


def _gml_info(graph_info, *args, **kwargs):
    ''' Format the graph information for the GML format '''
    info_str = "graph\n[\n"

    for key, val in iter(graph_info.items()):
        if not key.startswith("na_"):
            val = 1 if val is True else (0 if val is False else val)

            info_str += "  {} {}\n".format(key, val)

    return info_str


def _str_bytes_len(s):
    return len(s.encode('utf-8'))
